from __future__ import annotations

# Lazily handle stuff
import sys
import copy
import pathlib
import inspect
import functools
import importlib.util
from typing import Any, Dict, Callable, Union, Optional, Type, TypeVar, List, Tuple, cast, TYPE_CHECKING
from types import ModuleType

if TYPE_CHECKING:
    from lazyops.types import BaseModel
    from aiokeydb import KeyDBSession


# Lazy Importing
_imported_strings: Dict[str, Any] = {}


def import_string(dotted_path: str, is_module: Optional[bool] = None, allow_module: Optional[bool] = False) -> Any:
    """
    Taken from pydantic.utils.
    """
    from importlib import import_module
    if is_module:
        return import_module(dotted_path)
    try:
        module_path, class_name = dotted_path.strip(' ').rsplit('.', 1)
    except ValueError as e:
        raise ImportError(f'"{dotted_path}" doesn\'t look like a module path') from e

    module = import_module(module_path)
    try:
        return getattr(module, class_name)
    except AttributeError as e:
        if allow_module: return module
        raise ImportError(f'Module "{module_path}" does not define a "{class_name}" attribute') from e


def lazy_import(
    dotted_path: str,
    is_module: Optional[bool] = None,
    allow_module: Optional[bool] = False,
) -> Any:
    """
    Lazily imports a string with caching to avoid repeated imports
    """
    global _imported_strings
    if dotted_path not in _imported_strings:
        _imported_strings[dotted_path] = import_string(dotted_path, is_module = is_module, allow_module = allow_module)
    return _imported_strings[dotted_path]


@functools.lru_cache()
def import_function(func: Union[str, Callable]) -> Callable:
    """
    Imports a function from a string
    """
    return func if callable(func) else lazy_import(func)


def validate_callable(value: Optional[Union[str, Callable]]) -> Optional[Callable]:
    """
    Validates if the value is a callable
    """
    if value is None: return None
    return lazy_import(value) if isinstance(value, str) else value



def get_obj_class_name(obj: Any, is_parent: Optional[bool] = None) -> str:
    """
    Returns the module name + class name of an object

    args:
        obj: the object to get the class name of
        is_parent: if True, then it treats the object as unitialized and gets the class name of the parent
    """
    if is_parent is None: is_parent = inspect.isclass(obj)
    if is_parent: return f'{obj.__module__}.{obj.__name__}'
    return f'{obj.__class__.__module__}.{obj.__class__.__name__}'

def get_function_name(func: Callable) -> str:
    """
    Returns the module name + function name of a function
    """
    func = inspect.unwrap(func)
    return f'{func.__module__}.{func.__name__}'


def fetch_property(
    obj: Union[Type['BaseModel'], Dict],
    key: str,
    default: Optional[Any] = None
):  
    """
    Fetches a property from a dict or object
    """
    return obj.get(key, default) if isinstance(obj, dict) else getattr(obj, key, default)


def is_coro_func(obj, func_name: str = None):
    """
    This is probably in the library elsewhere but returns bool
    based on if the function is a coro
    """
    try:
        if inspect.iscoroutinefunction(obj): return True
        if inspect.isawaitable(obj): return True
        if func_name and hasattr(obj, func_name) and inspect.iscoroutinefunction(getattr(obj, func_name)):
            return True
        return bool(hasattr(obj, '__call__') and inspect.iscoroutinefunction(obj.__call__))

    except Exception:
        return False


def lazy_function(
    validator: Callable,
    function: Callable,
    *args,
    **kwargs,
):
    """
    Creates an empty function wrapper
    args:
        validator: function to validate the arguments
        func: function to call
        
    """
    def wrapper_func(func):
        if not validator():
            return func
        
        if is_coro_func(func):
            @functools.wraps(func)
            async def wrapped_func(*args, **kwargs):
                return await function(*args, **kwargs)
        
        else:
            @functools.wraps(func)
            def wrapped_func(*args, **kwargs):
                return function(*args, **kwargs)
        return wrapped_func
    return wrapper_func


_valid_class_init_kwarg: Dict[str, List[str]] = {}

def extract_obj_init_kwargs(obj: object) -> List[str]:
    """
    Extracts the kwargs that are valid for an object
    """
    global _valid_class_init_kwarg
    obj_name = get_obj_class_name(obj)
    if obj_name not in _valid_class_init_kwarg:
        argspec = inspect.getfullargspec(obj.__init__)
        _args = list(argspec.args)
        _args.extend(iter(argspec.kwonlyargs))
        # Check if subclass of Connection
        if hasattr(obj, "__bases__"):
            for base in obj.__bases__:
                _args.extend(extract_obj_init_kwargs(base))
        _valid_class_init_kwarg[obj_name] = list(set(_args))
    return _valid_class_init_kwarg[obj_name]



MT = TypeVar("MT", bound="BaseModel")

_extracted_base_model_kws: Dict[str, List[str]] = {}

def extract_base_model_kws(
    model: MT,
    kwargs: Dict[str, Any],
) -> Tuple[Dict[str, Any], Dict[str, Any]]:
    """
    Extracts the kwargs from the resource and returns the kwargs and the model kwargs
    """
    global _extracted_base_model_kws

    base_model_name = f"{model.__module__}.{model.__name__}"
    if base_model_name not in _extracted_base_model_kws:
        from lazyops.types.models import get_pyd_field_names
        resource_kws = get_pyd_field_names(model)
        # resource_kws = list(model.__fields__.keys())
        _extracted_base_model_kws[base_model_name] = resource_kws
    model_kwargs = {
        key: kwargs.pop(key) for key in kwargs if key in _extracted_base_model_kws[base_model_name]
    }
    return kwargs, model_kwargs


# Lazy KeyDB Sessions

_keydb_sessions: Dict[str, 'KeyDBSession'] = {}
_keydb_enabled: Optional[bool] = None

def get_keydb_enabled() -> bool:
    """
    Gets whether or not keydb is enabled
    """
    if _keydb_enabled is None:
        get_keydb_session(validate_active = True)
    return _keydb_enabled

def get_keydb_session(
    name: Optional[str] = "default",
    validate_active: Optional[bool] = None,
    **kwargs
) -> Optional["KeyDBSession"]:
    """
    Get the KeyDB Session
    """
    global _keydb_sessions, _keydb_enabled
    if not _keydb_sessions.get(name):
        from aiokeydb import KeyDBClient
        _keydb_sessions[name] = KeyDBClient.get_session(name = name, verbose = False, **kwargs)
    if validate_active and _keydb_enabled is None:
        try:
            _keydb_enabled = _keydb_sessions[name].ping()
        except Exception as e:
            from .logs import logger
            logger.warning(f"KeyDB Session {name} is not active: {e}")
            _keydb_enabled = False
    if validate_active and not _keydb_enabled:
        return None
    return _keydb_sessions[name]


"""
Lazily Handle Classes and Imports
"""

def import_code(
    code: str,
    add_to_sys: Optional[bool] = True,
    name: Optional[str] = None,
    ref: Optional[str] = None,
    module_name: Optional[str] = None,
) -> ModuleType:
    """
    Import dynamically generated code as a module. code is the
    object containing the code (a string, a file handle or an
    actual compiled code object, same types as accepted by an
    exec statement). The name is the name to give to the module,
    and the final argument says wheter to add it to sys.modules
    or not. If it is added, a subsequent import statement using
    name will return this module. If it is not added to sys.modules
    import will try to load it in the normal fashion. 
    """
    assert name or ref, 'Must provide either a name or a ref'
    name = name or ref
    module = ModuleType(name)
    module.__module__ = module_name or 'lazyops.types.dynamic'
    exec(code, module.__dict__)
    if add_to_sys:
        sys.modules[name] = module
    return module


DynamicT = TypeVar('DynamicT')
_DynamicClasses: Dict[str, DynamicT] = {}


def create_new_class(
    code: str,
    cls: DynamicT,
    ref: Optional[str] = None,
    module_name: Optional[str] = None,
) -> DynamicT:
    """
    Create a new class
    """
    global _DynamicClasses
    module_name = module_name or "lazyops.types.dynamic"
    from .serialization import create_hash_key
    code_hash = create_hash_key(kwargs = {'ref': ref, 'cls': cls, 'module_name': module_name, 'code': code})
    if code_hash in _DynamicClasses:
        return _DynamicClasses[code_hash]
    run_code = import_code(code, add_to_sys = True, ref = ref or cls.__name__, module_name = module_name)
    new_cls = copy.deepcopy(cls)
    for prop_or_func in dir(run_code):
        setattr(new_cls, prop_or_func, getattr(run_code, prop_or_func))
    _DynamicClasses[code_hash] = new_cls
    return new_cls


def load_module_from_file(
    file: Union[str, pathlib.Path], 
    module_name: str,
    add_to_sys: Optional[bool] = False,
) -> ModuleType:
    """
    Loads a module from a file
    """
    if isinstance(file, str):
        file = pathlib.Path(file)
    spec = importlib.util.spec_from_file_location(module_name, file.as_posix())
    module = importlib.util.module_from_spec(spec)
    if add_to_sys: sys.modules[module_name] = module
    spec.loader.exec_module(module)
    return module

def import_from_file(
    file: Union[str, pathlib.Path],
    cls_name: str,
) -> ModuleType:
    """
    Import a file
    """
    return load_module_from_file(file = file, module_name = cls_name)

def create_new_class_from_file(
    file: Union[str, pathlib.Path],
    cls_name: str,
    cls: DynamicT,
) -> DynamicT:
    """
    Import a file and create a new class
    """

    module_spec = load_module_from_file(file = file, module_name = cls_name)
    cls_spec = getattr(module_spec, cls_name)
    cls_spec = cast(cls, cls_spec)
    return cls_spec


from __future__ import annotations

"""
ThreadPool Implementation

- v3: Transform into a singleton proxy object class
- this is now imported from `lazyops.libs.pooler`
"""

from lazyops.libs.pooler import (
    set_concurrency_limit,
    get_concurrency_limit,
    is_coro_func,
    ThreadPool,
    ThreadPoolV2,
    ThreadPooler,
    ensure_coro,
)



from .models import (
    Params,
    BaseModel,
    Body,
    Field,
    Header,
    Depends,
    sjson_loads,
    sjson_dumps,
    jsonable_encoder,
)
from .exceptions import (
    BaseError,
    ErrorModel,
    ParseError,
    InvalidRequest,
    InternalError,
    MethodNotFound,
    InvalidParams,
    InternalError,
    NoContent
)
from .core import (
    JsonRpcRequest,
    JsonRpcResponse,
    JsonRpcContext,
    MethodRoute,
    RequestShadow,
    Entrypoint,
    EntrypointRoute,
    LazyJRPC,
)

__all__ = [
    'Params',
    'BaseModel',
    'Body',
    'Field',
    'Header',
    'Depends',
    'sjson_dumps',
    'sjson_loads',
    'jsonable_encoder',
    'BaseError',
    'ErrorModel',
    'ParseError',
    'InvalidRequest',
    'InternalError',
    'MethodNotFound',
    'InvalidParams',
    'InternalError',
    'NoContent',
    'JsonRpcRequest',
    'JsonRpcResponse',
    'JsonRpcContext',
    'MethodRoute',
    'RequestShadow',
    'Entrypoint',
    'EntrypointRoute',
    'LazyJRPC',
]



import abc
from typing import Union
from lazyops.utils.pooler import ThreadPoolV2 as ThreadPooler

class BaseCompress(abc.ABC):

    @abc.abstractclassmethod
    def compress(cls, data: Union[str, bytes], **kwargs) -> bytes:
        """
        Base Compress
        """
        pass

    @abc.abstractclassmethod
    def decompress(cls, data: Union[str, bytes], **kwargs) -> bytes:
        """
        Base Decompress
        """
        pass


    @abc.abstractclassmethod
    async def acompress(cls, data: Union[str, bytes], **kwargs) -> bytes:
        """
        Base Compress
        """
        return await ThreadPooler.run_async(cls.compress, data, **kwargs)
    
    @abc.abstractclassmethod
    async def adecompress(cls, data: Union[str, bytes], **kwargs) -> bytes:
        """
        Base Decompress
        """
        return await ThreadPooler.run_async(cls.decompress, data, **kwargs)
    




from typing import List
from .types import AsyncTask, TaskStatus, WaitType, ErrorHandling
from .errors import AsyncTaskException


def await_all(
    tasks: List[AsyncTask], 
    error_handling: ErrorHandling = ErrorHandling.RAISE
):
    """
    Await all tasks to complete

    :param tasks: List of tasks
    :param error_handling: Error handling strategy (raise or ignore)
    """
    result = []
    for task in tasks:
        if task.status == TaskStatus.PENDING:
            while not task.thread.is_alive(): pass
            task.thread.join()

    for task in tasks:
        if task.status == TaskStatus.SUCCESS:
            result.append(task.result)
        elif task.status == TaskStatus.FAILURE:
            if error_handling == ErrorHandling.RAISE:
                raise AsyncTaskException(str(task.exception), task.func.__name__)
    
    return result


def await_first(
    tasks: List[AsyncTask],
    error_handling: ErrorHandling = ErrorHandling.RAISE
):
    """
    Await first task to complete

    :param tasks: List of tasks
    :param error_handling: Error handling strategy (raise or ignore)
    """

    first_failed: AsyncTask = None
    while tasks:
        for i in range(len(tasks)):
            if tasks[i].status == TaskStatus.SUCCESS:
                return tasks[i].result
            elif tasks[i].status == TaskStatus.FAILURE:
                if first_failed is None: first_failed = tasks[i]
                tasks.pop(i)
                i -= 1
    if first_failed is not None and error_handling == ErrorHandling.RAISE:
        raise AsyncTaskException(str(first_failed.exception), first_failed.func.__name__)
    return None

def await_until(
    tasks: List[AsyncTask],
    wait_type: WaitType = WaitType.ALL,
    error_handling: ErrorHandling = ErrorHandling.RAISE
):
    """
    Await until tasks are completed based on wait_type

    :param tasks: List of tasks
    :param wait_type: Wait type (all or first)
    :param error_handling: Error handling strategy (raise or ignore)
    """
    if wait_type == WaitType.ALL:
        return await_all(tasks, error_handling)
    elif wait_type == WaitType.FIRST:
        return await_first(tasks, error_handling)
    else:
        raise ValueError(f"Invalid wait_type: {wait_type}")



import os
import pathlib

from typing import Any, Type, Tuple, Dict, List, Union, Optional, Callable, TypeVar, Generic, TYPE_CHECKING
from lazyops.types.formatting import to_camel_case, to_snake_case, to_graphql_format
from lazyops.types.classprops import classproperty, lazyproperty
from lazyops.utils.serialization import Json

from pydantic import Field
from pydantic.networks import AnyUrl
from lazyops.imports._pydantic import BaseSettings as _BaseSettings
from lazyops.imports._pydantic import BaseModel as _BaseModel
from lazyops.imports._pydantic import (
    validator, 
    root_validator, 
    pre_root_validator, 
    get_pyd_dict, 
    get_pyd_fields, 
    get_pyd_field_names,
    get_pyd_fields_dict,
    get_pyd_schema,
    pyd_parse_obj,
    ConfigDict,
    PYD_VERSION,
)

"""
Migration to pydantic v2 with backwards compatibility
"""


def schema_extra(schema: dict, _):
    props = {
        k: v
        for k, v in schema.get('properties', {}).items()
        if not v.get("hidden", False) and not k.startswith('_')
    }
    schema["properties"] = props


class BaseSettings(_BaseSettings):

    class Config:
        env_prefix: str = ""
        case_sensitive: bool = False


    def update_config(self, **kwargs):
        """
        Update the config for the other settings
        """
        for k, v in kwargs.items():
            if not hasattr(self, k): continue
            if isinstance(getattr(self, k), pathlib.Path):
                setattr(self, k, pathlib.Path(v))
            elif isinstance(getattr(self, k), self.__class__):
                val = getattr(self, k)
                if hasattr(val, 'update_config'):
                    val.update_config(**v)
                else:
                    val = val.__class__(**v)
                setattr(self, k, val)
            else: 
                setattr(self, k, v)

    def set_env(self):
        """
        Update the Env variables for the current session
        """
        data = get_pyd_dict(self, exclude_none=True)
        for k, v in data.items():
            if isinstance(v, BaseSettings):
                v.set_env()
            else:
                os.environ[self.Config.env_prefix + k.upper()] = str(v)
    

class ProxySettings:
    def __init__(
        self,
        settings_cls: Optional[Type[BaseSettings]] = None,
        settings_getter: Optional[Union[Callable, str]] = None,
        debug_enabled: Optional[bool] = False,
    ):
        """
        Proxy settings object
        """
        # Intentionally underscore on the end to avoid conflicts with the settings
        assert settings_cls or settings_getter, "Either settings_cls or settings_getter must be provided"
        self.__settings_cls_ = settings_cls
        self.__settings_getter_ = settings_getter
        if self.__settings_getter_ and isinstance(self.__settings_getter_, str):
            from lazyops.utils.helpers import import_string
            self.__settings_getter_ = import_string(self.__settings_getter_)
        self.__settings_ = None
        self.__debug_enabled_ = debug_enabled
        self.__last_attrs_: Dict[str, int] = {}

    @property
    def _settings_(self):
        """
        Returns the settings object
        """
        if self.__settings_ is None:
            if self.__settings_getter_:
                self.__settings_ = self.__settings_getter_()
            elif self.__settings_cls_:
                self.__settings_ = self.__settings_cls_()
        return self.__settings_

    def __getattr__(self, name):
        """
        Forward all unknown attributes to the settings object
        """
        if not self.__debug_enabled_:
            return getattr(self._settings_, name)
        
        # Try to debug the attribute
        if name not in self.__last_attrs_:
            self.__last_attrs_[name] = 0
        self.__last_attrs_[name] += 1
        if self.__last_attrs_[name] > 5:
            raise AttributeError(f"Settings object has no attribute {name}")

        if hasattr(self._settings_, name):
            self.__last_attrs_[name] = 0
            return getattr(self._settings_, name)
        raise AttributeError(f"Settings object has no attribute {name}")


class BaseModel(_BaseModel):

    if PYD_VERSION == 1:
        class Config:
            extra = 'allow'
            arbitrary_types_allowed = True
            alias_generator = to_snake_case
    else:
        model_config = ConfigDict(extra = 'allow', arbitrary_types_allowed = True)
        
                
    def get(self, name, default: Any = None):
        """
        Get an attribute from the model
        """
        return getattr(self, name, default)
    
    def update(self, **kwargs):
        """
        Update the model attributes
        """
        for k, v in kwargs.items():
            if not hasattr(self, k): continue
            setattr(self, k, v)


    @classmethod
    def create_one(
        cls,
        **kwargs
    ) -> Tuple[Type['BaseModel'], Dict]:
        """
        Extracts the resource from the kwargs and returns the resource 
        and the remaining kwargs
        """
        resource_fields = get_pyd_field_names(cls)
        # resource_fields = [field.name for field in cls.__fields__.values()]

        resource_kwargs = {k: v for k, v in kwargs.items() if k in resource_fields}
        return_kwargs = {k: v for k, v in kwargs.items() if k not in resource_fields}
        resource_obj = pyd_parse_obj(cls, resource_kwargs)
        # resource_obj = cls.parse_obj(resource_kwargs)
        return resource_obj, return_kwargs
    
    @classmethod
    def create_many(cls, data: List[Dict]) -> List['BaseModel']:
        """
        Create a list of resources from a list of dicts
        """
        return [pyd_parse_obj(cls, d) for d in data]
        # return [cls.parse_obj(d) for d in data]
    

    # @classproperty
    # def model_field_names(cls) -> List[str]:
    #     """
    #     Returns the model fields names
    #     """
    #     return get_pyd_field_names(cls)
        # return [field.name for field in cls.__fields__.values()]
    
    @classmethod
    def get_model_field_names(cls) -> List[str]:
        """
        Get the model fields
        """
        return get_pyd_field_names(cls)
        # return [field.name for field in self.__fields__.values()]

    def replace(self, obj: Type['BaseModel']):
        """Replace current attributes with `obj` attributes."""
        for field in obj.get_model_field_names():
            if hasattr(self, field):
                setattr(self, field, getattr(obj, field))
    
    def graphql(
        self,
        *,
        include: Optional[Any] = None,
        exclude: Optional[Any] = None,
        by_alias: bool = False,
        skip_defaults: Optional[bool] = None,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
    ) -> Union[Dict, Any]:
        """
        Return the data in `graphql` format
        """
        data = get_pyd_dict(
            include = include, exclude = exclude, \
            by_alias = by_alias, skip_defaults = skip_defaults, \
            exclude_unset = exclude_unset, exclude_defaults = exclude_defaults, \
            exclude_none = exclude_none
        )

        # data = self.dict(
        #     include = include, exclude = exclude, \
        #     by_alias = by_alias, skip_defaults = skip_defaults, \
        #     exclude_unset = exclude_unset, exclude_defaults = exclude_defaults, \
        #     exclude_none = exclude_none
        # )
        return to_graphql_format(data)

    @classmethod
    def parse_obj(
        cls,
        obj: Any,
        strict: Optional[bool] = False,
        from_attributes: Optional[bool] = True,
        **kwargs
    ) -> 'BaseModel':
        """
        Parses an object into the resource
        """
        return pyd_parse_obj(cls, obj, strict = strict, from_attributes = from_attributes, **kwargs)
    
    def dict(self, **kwargs):
        """
        Returns the dict representation of the response
        """
        return get_pyd_dict(self, **kwargs)

    def schema(self, **kwargs):
        """
        Returns the dict representation of the response
        """
        return get_pyd_schema(self, **kwargs)



class Schema(BaseModel):

    class Config:
        extra = 'allow'
        arbitrary_types_allowed = True
        if PYD_VERSION == 1:
            json_dumps = Json.dumps
        # alias_generator = to_camel_case

    # @lazyproperty
    # def schema_fields(self) -> List[str]:
    #     return [field.name for field in self.__fields__.values()]

    def get(self, name, default: Any = None):
        return getattr(self, name, default)


class Constant(tuple):
    "Pretty display of immutable constant."

    def __new__(cls, name):
        return tuple.__new__(cls, (name,))

    def __repr__(self):
        return f'{self[0]}'


class UrlModel(BaseModel):
    """
    Parse a URL into its components
    """
    url: AnyUrl
    scheme: Optional[str] = 'http'
    adapter: Optional[str] = None

    @lazyproperty
    def db(self):
        return self.url.path[1:]
    
    @lazyproperty
    def path(self):
        return self.url.path

    @lazyproperty
    def host(self):
        return self.url.host

    @lazyproperty
    def port(self):
        return self.url.port

    @lazyproperty
    def user(self):
        return self.url.user or None
    
    @lazyproperty
    def password(self):
        return self.url.password or None
    
    @lazyproperty
    def api_key(self):
        return self.url.user or None
    
    @lazyproperty
    def token(self):
        return self.url.password or None

    @lazyproperty
    def server_uri(self):
        """
        Returns the bare connection uri without the database/path/auth
        """
        return self.get_uri(include_auth = False, include_db = False, include_path = False)
    
    @lazyproperty
    def base_uri(self):
        """
        Returns the connection uri without the database/path
        """
        return self.get_uri(include_path = False, include_db = False)
    
    @lazyproperty
    def safe_uri(self):
        """
        Returns the connection uri without the password
        """
        return self.get_uri(include_auth = True, include_db = True, include_path = True, include_adapter = True, _safe = True)

    @lazyproperty
    def connection_uri(self):
        """
        Returns the connection uri with the database/path
        """
        return self.get_uri(include_path = True, include_db = True, include_auth = True, include_adapter = True, _safe = False)

    def get_uri(
        self, 
        scheme: Optional[str] = None, 
        db: Optional[str] = None,
        path: Optional[str] = None,
        include_path: Optional[bool] = True,
        include_db: Optional[bool] = True,
        include_auth: Optional[bool] = True,
        include_adapter: Optional[bool] = True,
        _safe: Optional[bool] = False,
        **kwargs,
    ):
        scheme = scheme or self.scheme
        if include_adapter and self.adapter and self.adapter not in scheme:
            scheme = f'{self.adapter}+{scheme}'
        uri = f'{scheme}://'
        if include_auth and self.user:
            uri += '*****' if _safe else f'{self.user}'
            if self.password:
                uri += ':*****' if _safe else f':{self.password}'
            uri += '@'
        uri += f'{self.host}'
        if self.port:
            uri += f':{self.port}'
        if 'http' in scheme and include_path:
            path = path or self.path
            if path: uri += f'/{path}'
        elif include_db:
            db = db or self.db
            if db: uri += f'/{db}'
        return uri

    @classmethod
    def parse(
        cls, 
        url: Optional[str] = None, 
        scheme: Optional[str] = 'http',
        adapter: Optional[str] = None,
        **config
    ):
        path = config.get('path')
        db = config.get('db')
        if url: 
            if 'http' in scheme and path: url += f'/{path}'
            elif db: url += f'/{db}'    
            return cls(url = url, scheme=scheme, adapter=adapter)
        url = f'{scheme}://'
        host = config.get('host')
        port = config.get('port')
        user = config.get('user')
        password = config.get('password')
        api_key = config.get('api_key')
        token = config.get('token')
        
        if user:
            url += f'{user}'
            if password:
                url += f':{password}'
            url += '@'
        elif api_key:
            url += f'{api_key}@'
        elif token:
            url += f'{token}@'
        url += f'{host}'
        if port: url += f':{port}'
        if 'http' in scheme and path: url += f'/{path}'
        elif db: url += f'/{db}'
        return cls(url = url, scheme = scheme, adapter = adapter)

class BaseDBUrl(UrlModel):

    @classmethod
    def parse(
        cls, 
        url: Optional[str] = None, 
        scheme: Optional[str] = 'http',
        adapter: Optional[str] = None,
        **config
    ):
        config_key = cls.__name__.split('DB', 1)[0].lower()
        db = config.get(f'{config_key}_db', config.get('db'))
        path = config.get(f'{config_key}_path', config.get('path'))
        
        if url: 
            if 'http' in scheme and path: url += f'/{path}'
            elif db: url += f'/{db}'    
            return cls(url = url, scheme=scheme, adapter=adapter)

        
        url = f'{scheme}://'
        host = config.get(f'{config_key}_host', config.get('host'))
        port = config.get(f'{config_key}_port', config.get('port'))
        user = config.get(f'{config_key}_user', config.get('user'))
        password = config.get(f'{config_key}_password', config.get('password'))
        api_key = config.get(f'{config_key}_api_key', config.get('api_key'))
        token = config.get(f'{config_key}_token', config.get('token'))
        
        if user:
            url += f'{user}'
            if password: url += f':{password}'
            url += '@'
        elif api_key:
            url += f'{api_key}@'
        elif token:
            url += f'{token}@'
        url += f'{host}'
        if port:
            url += f':{port}'
        if 'http' in scheme and path: url += f'/{path}'
        elif db: url += f'/{db}'
        return cls(url = url, scheme = scheme, adapter = adapter)


class PostgresDB(BaseDBUrl):
    url: AnyUrl
    scheme: Optional[str] = 'postgresql'
    adapter: Optional[str] = 'postgresql'
    
    @classmethod
    def parse(
        cls, 
        url: Optional[str] = None, 
        scheme: Optional[str] = 'postgresql',
        adapter: Optional[str] = 'postgresql',
        **config
    ):
        return super().parse(url = url, scheme = scheme, adapter = adapter, **config)

class RedisDB(BaseDBUrl):
    url: AnyUrl
    scheme: Optional[str] = 'redis'
    adapter: Optional[str] = None
    
    @classmethod
    def parse(
        cls, 
        url: Optional[str] = None, 
        scheme: Optional[str] = 'redis',
        adapter: Optional[str] = None,
        **config
    ):
        return super().parse(url = url, scheme = scheme, adapter = adapter, **config)

class NatsDB(BaseDBUrl):
    url: AnyUrl
    scheme: Optional[str] = 'nats'
    adapter: Optional[str] = None
    
    @classmethod
    def parse(
        cls, 
        url: Optional[str] = None, 
        scheme: Optional[str] = 'nats',
        adapter: Optional[str] = None,
        **config
    ):
        return super().parse(url = url, scheme = scheme, adapter = adapter, **config)


        


__all__ = [
    'Field',
    'BaseModel',
    'BaseSettings',
    'Schema',
    'validator',
    'UrlModel',
    'BaseDBUrl',
    'PostgresDB',
    'RedisDB',
    'NatsDB',
]


import random
import asyncio
import pathlib

from pydantic import BaseModel
from lazyops.libs.persistence import PersistentDict
from lazyops.libs.persistence.main import _DEBUG_ENABLED

_DEBUG_ENABLED = True

class X(BaseModel):
    n: int


def test_json(backend: str = 'local', compression_level: int = None, **kwargs):
    """
    Test JSON - The object is serialized to a dict, and upon retrieval, the type is lost
    """
    
    n = random.randint(0, 1000)
    d = PersistentDict(
        name = 'test',
        serializer = 'json',
        backend_type = backend,
        base_key = 'test.json',
        serializer_kwargs= {
            'compression_level': compression_level,
            'disable_object_serialization': True,
        },
        **kwargs,
    )

    d['x'] = X(n=n)
    print(f'[{backend} - {compression_level} - {kwargs} - {d.base_key}] test json: {n}', d['x'], ' type: ', type(d['x']))
    assert d['x']['n'] == n, f'[{backend} - {compression_level} - {kwargs}] test json: {n} {d["x"]} {type(d["x"])}'

    del d['x']
    assert 'x' not in d


def test_json_obj(backend: str = 'local', compression_level: int = None, **kwargs):
    """
    Test JSON Object - The object is serialized to a dict, however upon retrieval, the object gets re-instantiated
    due to the serialization_obj parameter
    """
    
    n = random.randint(0, 1000)
    d = PersistentDict(
        name = 'test',
        serializer = 'json',
        backend_type = backend,
        base_key = 'test.json.obj',
        serializer_kwargs = {
            'serialization_obj': X,
            'compression_level': compression_level,
        },
        **kwargs,
    )

    d['x'] = X(n=n)

    print(f'[{backend} - {compression_level} - {kwargs} - {d.base_key}] test json obj: {n} ', d['x'], ' type: ', type(d['x']))
    assert d['x'].n == n

    del d['x']
    assert 'x' not in d


def test_msgpack(backend: str = 'local', compression_level: int = None, **kwargs):
    """
    Test MsgPack Object - The object is serialized to a dict, however upon retrieval, the object gets re-instantiated
    due to the serialization_obj parameter
    """
    
    n = random.randint(0, 1000)
    d = PersistentDict(
        name = 'test',
        serializer = 'msgpack',
        backend_type = backend,
        base_key = 'test.msgpack',
        serializer_kwargs = {
            'serialization_obj': X,
            'compression_level': compression_level,
        },
        **kwargs,
    )

    d['x'] = X(n=n)
    print(f'[{backend} - {compression_level} - {kwargs} - {d.base_key}] test msgpack obj: {n} ', d['x'], ' type: ', type(d['x']))
    assert d['x'].n == n

    del d['x']
    assert 'x' not in d

def test_pickle(backend: str = 'local', compression_level: int = None, **kwargs):
    """
    Pickle Test - The object is pickled and thus the type is perserved
    """
    n = random.randint(0, 1000)
    d = PersistentDict(
        name = 'test',
        serializer = 'pickle',
        backend_type = backend,
        base_key = 'test.pickle',
        serializer_kwargs= {
            'compression_level': compression_level,
        },
        **kwargs,

    )

    d['x'] = X(n=n)
    print(f'[{backend} - {compression_level} - {kwargs} - {d.base_key}] test pickle: {n} ', d['x'], ' type: ', type(d['x']))
    assert d['x'].n == n
    del d['x']
    assert 'x' not in d


def test_basic_local(compression_level: int = None, **kwargs):
    """
    Test Basic Functionality
    """

    test_json(backend='local', compression_level=compression_level, **kwargs)
    test_json_obj(backend='local', compression_level=compression_level, **kwargs)
    test_msgpack(backend='local', compression_level=compression_level, **kwargs)
    test_pickle(backend='local', compression_level=compression_level, **kwargs)


def test_basic(compression_level: int = None, **kwargs):
    """
    Test Basic Functionality
    """

    test_json(backend='redis', compression_level=compression_level, **kwargs)
    test_json_obj(backend='redis', compression_level=compression_level, **kwargs)
    test_msgpack(backend='redis', compression_level=compression_level, **kwargs)
    test_pickle(backend='redis', compression_level=compression_level, **kwargs)

    test_json(backend='auto', compression_level=compression_level, **kwargs)
    test_pickle(backend='auto', compression_level=compression_level, **kwargs)

async def atest_json(backend: str = 'local', compression_level: int = None, **kwargs):
    n = random.randint(0, 1000)
    d = PersistentDict(
        name = 'test',
        serializer = 'json',
        backend_type = backend,
        base_key = 'test.json',
        serializer_kwargs= {
            'compression_level': compression_level,
            'disable_object_serialization': True,
        },
        **kwargs,
    )
    await d.aset('x', X(n=n))
    v = await d.aget('x')

    print(f'[{backend} - {compression_level} - {kwargs} - {d.base_key}] async test json: {n} ', v, ' type: ', type(v))
    assert v['n'] == n
    await d.adelete('x')
    assert 'x' not in d


async def atest_json_obj(backend: str = 'local', compression_level: int = None, **kwargs):
    n = random.randint(0, 1000)
    d = PersistentDict(
        name = 'test',
        serializer = 'json',
        backend_type = backend,
        base_key = 'test.json.obj',
        serializer_kwargs = {
            'serialization_obj': X,
            'compression_level': compression_level,
            
        }, 
        **kwargs
    )

    await d.aset('x', X(n=n))

    v = await d.aget('x')
    print(f'[{backend} - {compression_level} - {kwargs} - {d.base_key}] async test json obj: {n} ', v, ' type: ', type(v))
    assert v.n == n

    await d.adelete('x')
    assert 'x' not in d

async def atest_pickle(backend: str = 'local', compression_level: int = None, **kwargs):

    n = random.randint(0, 1000)
    d = PersistentDict(
        name = 'test',
        serializer = 'pickle',
        backend_type = backend,
        base_key = 'test.pickle',
        serializer_kwargs= {
            'compression_level': compression_level,
        },
        **kwargs,
    )

    await d.aset('x', X(n=n))
    v = await d.aget('x')
    print(f'[{backend} - {compression_level} - {kwargs} - {d.base_key}] async test pickle: {n} ', v, ' type: ', type(v))
    assert v.n == n
    await d.adelete('x')
    assert 'x' not in d

async def atest_basic_local(compression_level: int = None, **kwargs):
    """
    Test Basic Functionality
    """

    await atest_json(backend='local', compression_level=compression_level, **kwargs)
    await atest_json_obj(backend='local', compression_level=compression_level, **kwargs)
    await atest_pickle(backend='local', compression_level=compression_level, **kwargs)

async def atest_basic(compression_level: int = None, **kwargs):
    """
    Test Basic Functionality
    """

    await atest_json(backend='redis', compression_level=compression_level, **kwargs)
    await atest_json_obj(backend='redis', compression_level=compression_level, **kwargs)

    await atest_pickle(backend='redis', compression_level=compression_level, **kwargs)

    await atest_json(backend='auto', compression_level=compression_level, **kwargs)
    await atest_pickle(backend='auto', compression_level=compression_level, **kwargs)


async def run_basic_tests():

    fp = pathlib.Path(__file__).parent
    test_basic()
    await atest_basic()

    test_basic_local(file_path = fp)
    await atest_basic_local(file_path = fp)

    test_basic(compression_level=1)
    await atest_basic(compression_level=1)

    test_basic(compression_level=9)
    await atest_basic(compression_level=9)

    test_basic(compression_level=1, hset_disabled=True)
    await atest_basic(compression_level=1, hset_disabled=True)

"""
Advanced Tests
"""


def test_json_advanced(backend: str = 'local', compression: str = None, compression_level: int = None, is_obj: bool = False, **kwargs):
    """
    Handling more advanced dict types
    """
    d = PersistentDict(
        name='test',
        serializer='json',
        backend_type=backend,
        base_key='test.json',
        serializer_kwargs={
            'compression': compression,
            'compression_level': compression_level,
            'disable_object_serialization': not is_obj,
            'serialization_obj': X if is_obj else None,
        },
        **kwargs
    )
    y = {}
    for i in range(10):
        i_int = random.randint(0, 1000)
        y[f'x{i}'] = i_int
        d[f'x{i}'] = X(n = i_int)
    
    d.setdefault('xy', {})
    d['xy'].update({
        'x': 1,
        'y': 2,
    })
    
    print(f'[{backend} - {compression}:{d.compression_level} - {kwargs} - {d.base_key}] test json advanced: ', y, '=', list(d.keys()))
    for k,v in d.items():
        if k == 'xy':
            print(f'[{backend} - {compression}:{d.compression_level} - {kwargs} - {d.base_key}] test json advanced: {k} {v} {type(v)}')
            continue
        assert y[k] == v.n if is_obj else v['n'], f'[{backend} - {compression}:{d.compression_level} - {kwargs}] test json advanced: {k} {y[k]} {v} {type(v)}'
    d.clear()
    assert len(d) == 0


def test_msgpack_advanced(backend: str = 'local', compression: str = None, compression_level: int = None, **kwargs):
    """
    Test MsgPack Object - The object is serialized to a dict, however upon retrieval, the object gets re-instantiated
    due to the serialization_obj parameter
    """
    d = PersistentDict(
        name='test',
        serializer='msgpack',
        backend_type=backend,
        base_key='test.msgpack',
        serializer_kwargs={
            'serialization_obj': X,
            'compression': compression,
            'compression_level': compression_level,
        },
        **kwargs
    )
    y = {}
    for i in range(10):
        i_int = random.randint(0, 1000)
        y[f'x{i}'] = i_int
        d[f'x{i}'] = X(n = i_int)
    

    d.setdefault('xy', {})
    d['xy'].update({
        'x': 1,
        'y': 2,
    })
    

    print(f'[{backend} - {d.base_key}] mutating: ', d['x5'].n, '=', d['x5'].n + 1)
    d['x5'].n += 1
    print(f'[{backend} - {d.base_key}] mutated: ', d['x5'].n)
    
    d['x5'].n += 1
    print(f'[{backend} - {d.base_key}] mutated: ', d['x5'].n)

    d['x5'].n *= 5

    y['x5'] += 2
    y['x5'] *= 5

    print(f'[{backend} - {compression}:{compression_level} - {kwargs} - {d.base_key}] test msgpack advanced: ', y, '=', list(d.keys()))
    for k,v in d.items():
        if k == 'xy':
            print(f'[{backend} - {compression}:{compression_level} - {kwargs} - {d.base_key}] test msgpack advanced: {k} {v} {type(v)}')
            continue
        assert y[k] == v.n, f'[{backend} - {compression}:{compression_level} - {kwargs}] test msgpack advanced: {k} {y[k]} {v} {type(v)}'
    d.clear()
    assert len(d) == 0


def test_pickle_advanced(backend: str = 'local', compression: str = None, compression_level: int = None, **kwargs):
    """
    Pickle Test - The object is pickled and thus the type is perserved
    """
    d = PersistentDict(
        name='test',
        serializer='pickle',
        backend_type=backend,
        base_key='test.pickle',
        serializer_kwargs={
            'compression': compression,
            'compression_level': compression_level,
        },
        **kwargs
    )
    y = {}
    for i in range(10):
        i_int = random.randint(0, 1000)
        y[f'x{i}'] = i_int
        d[f'x{i}'] = X(n = i_int)
    
    d.setdefault('xy', {})
    d['xy'].update({
        'x': 1,
        'y': 2,
    })

    print(f'[{backend} - {d.base_key}] mutating: ', d['x5'].n, '=', d['x5'].n + 1)
    d['x5'].n += 1
    print(f'[{backend} - {d.base_key}] mutated: ', d['x5'].n)
    
    d['x5'].n += 1
    print(f'[{backend} - {d.base_key}] mutated: ', d['x5'].n)

    d['x5'].n *= 5
    # print(f'[{backend} - {d.base_key}] mutated: ', d['x5'].n)

    y['x5'] += 2
    y['x5'] *= 5

    print(f'[{backend} - {compression}:{d.compression_level} - {kwargs} - {d.base_key}] test pickle advanced: ', y, '=', list(d.keys()))
    for k,v in d.items():
        if k == 'xy':
            print(f'[{backend} - {compression}:{d.compression_level} - {kwargs} - {d.base_key}] test pickle advanced: {k} {v} {type(v)}')
            continue

        assert y[k] == v.n, f'[{backend} - {compression}:{d.compression_level} - {kwargs}] test pickle advanced: {k} {y[k]} {v} {type(v)}'
    for key in d:
        print(f'[{backend} - {compression}:{d.compression_level} - {kwargs} - {d.base_key}] test pickle advanced: {key} {d[key]} {type(d[key])}')
    d.clear()
    assert len(d) == 0

def test_basic_advanced(compression_level: int = None, **kwargs):
    """
    Test Basic Functionality
    """
    fp = pathlib.Path(__file__).parent

    for compression in {
        None,
        'gzip',
        'zlib',
        'lz4',
        'zstd',
    }:

        test_json_advanced(backend='local', compression=compression, compression_level=compression_level, file_path = fp, **kwargs)
        test_msgpack_advanced(backend='local',compression=compression, compression_level=compression_level, file_path = fp, **kwargs)
        test_pickle_advanced(backend='local', compression=compression, compression_level=compression_level, file_path = fp, **kwargs)


        test_json_advanced(backend='redis', compression=compression, compression_level=compression_level, **kwargs)
        # test_msgpack_advanced(backend='redis', compression=compression, compression_level=compression_level, **kwargs)
        test_pickle_advanced(backend='redis', compression=compression, compression_level=compression_level, **kwargs)

        test_json_advanced(backend='auto', compression=compression, compression_level=compression_level, **kwargs)
        test_pickle_advanced(backend='auto', compression=compression, compression_level=compression_level, **kwargs)



async def entrypoint():
    # await run_basic_tests()
    test_basic_advanced()



if __name__ == '__main__':
    asyncio.run(entrypoint())




import time
import dateparser

from datetime import datetime, timedelta, timezone
from dataclasses import dataclass
from typing import Any, List, Optional, Union, Callable

@dataclass
class LazyData:
    string: str = None
    value: Any = None
    dtype: str = None



TimeValues = {
    'secs': 60,
    'mins': 60,
    'hrs': 60,
    'days': 24,
    'wks': 7,
    'mnths': 4
}

class TimeItem:
    secs: float

    @property
    def mins(self):
        return self.secs / 60
    
    @property
    def hrs(self):
        return self.mins / 60
    
    @property
    def days(self):
        return self.hrs / 24
    
    @property
    def wks(self):
        return self.days / 7
    
    @property
    def mnths(self):
        return self.wks / 4
    
    @property
    def ablstime(self):
        curr_val = self.secs
        dict_val, str_val = {}, ''
        for tkey, tnum in TimeValues.items():
            if curr_val >= 1:
                tval = tkey[0] if self.short else tkey
                curr_val, curr_num = divmod(curr_val, tnum)
                if type(curr_num) == float:
                    str_val = f'{curr_num:.1f} {tval} ' + str_val
                else:
                    str_val = f'{curr_num} {tval} ' + str_val
                dict_val[tkey] = curr_num
        str_val = str_val.strip()
        return LazyData(string=str_val, value=dict_val, dtype='time')



class LazyTime:
    def __init__(self, t=None, short=False, *args, **kwargs):
        self.t = time.time() if not t else t
        self.short = short
        self._now = None

    @property
    def now(self):
        if self._now:
            return self._now
        return time.time()

    def stop(self):
        self._now = time.time()

    @property
    def diff(self):
        return self.now - self.t

    @property
    def secs(self):
        return self.diff

    @property
    def mins(self):
        return self.secs / 60
    
    @property
    def hrs(self):
        return self.mins / 60
    
    @property
    def days(self):
        return self.hrs / 24
    
    @property
    def wks(self):
        return self.days / 7
    
    @property
    def mnths(self):
        return self.wks / 4
    
    @property
    def ablstime(self):
        curr_val = self.secs
        dict_val, str_val = {}, ''
        for tkey, tnum in TimeValues.items():
            if curr_val >= 1:
                tval = tkey[0] if self.short else tkey
                curr_val, curr_num = divmod(curr_val, tnum)
                if type(curr_num) == float:
                    str_val = f'{curr_num:.1f} {tval} ' + str_val
                else:
                    str_val = f'{curr_num} {tval} ' + str_val
                dict_val[tkey] = curr_num
        str_val = str_val.strip()
        return LazyData(string=str_val, value=dict_val, dtype='time')
        
    @property
    def s(self):
        return self.secs
    
    @property
    def seconds(self):
        return self.secs

    @property
    def m(self):
        return self.mins
    
    @property
    def minutes(self):
        return self.mins
    
    @property
    def h(self):
        return self.hrs
    
    @property
    def hr(self):
        return self.hrs
    
    @property
    def hour(self):
        return self.hrs
    
    @property
    def hours(self):
        return self.hrs

    @property
    def d(self):
        return self.days
    
    @property
    def day(self):
        return self.days
    
    @property
    def w(self):
        return self.wks
    
    @property
    def wk(self):
        return self.wks
    
    @property
    def week(self):
        return self.wks

    @property
    def weeks(self):
        return self.wks
    
    @property
    def month(self):
        return self.mnths
    
    @property
    def mons(self):
        return self.mnths
    
    @property
    def months(self):
        return self.mnths



class LazyDate:
    @classmethod
    def dtime(cls, dt: str = None, prefer: str = 'past'):
        if not dt:
            return datetime.now(timezone.utc).isoformat('T')
        pdt = dateparser.parse(dt, settings={'PREFER_DATES_FROM': prefer, 'TIMEZONE': 'UTC', 'RETURN_AS_TIMEZONE_AWARE': True})
        if prefer == 'past': 
            r = (datetime.now(timezone.utc) - pdt)
        else:
            r = (pdt - datetime.now(timezone.utc))
        return TimeItem(r.total_seconds())

    @classmethod
    def date(cls, dt: str = None, prefer: str = 'past'):
        if not dt:
            return datetime.now(timezone.utc)
        return dateparser.parse(dt, settings={'PREFER_DATES_FROM': prefer, 'TIMEZONE': 'UTC', 'RETURN_AS_TIMEZONE_AWARE': True})    

    @classmethod
    def fmt(cls, secs, **kwargs):
        return TimeItem(secs)
    
    @property
    def now(self):
        return self.dtime()

    def __call__(self, **kwargs):
        return self.dtime(**kwargs)

class LazyFormatter:
    def size(cls, bvalue, suffix="B"):
        factor = 1024
        for unit in ["", "K", "M", "G", "T", "P"]:
            if bvalue < factor:
                return LazyData(string=f"{bvalue:.2f} {unit}{suffix}", value=bvalue, dtype="size")
            bvalue /= factor

    def ftime(cls, t: time.time = None, short: bool = True):
        return LazyTime(t, short=short)
    
    def dtime(cls, dt: str = None, prefer: str = 'past'):
        return LazyDate.dtime(dt, prefer)


class LazyTimer:
    timers = {}
    def stop_timer(self, name):
        assert name in LazyTimer.timers
        LazyTimer.timers[name].stop()
        return LazyTimer.timers[name]

    def __getitem__(self, name, t = None, *args, **kwargs):
        return LazyTimer.timers.get(name, None)

    def __call__(self, name, t = None, *args, **kwargs):
        if name not in LazyTimer.timers:
            LazyTimer.timers[name] = LazyTime(t=t, *args, **kwargs)
        return LazyTimer.timers[name]


class LazyObject(dict):
    def __init__(self, data, **kwargs):
        for name, val in data.items():
            self[name] = val
        for name, val in kwargs.items():
            self[name] = val

    def getorset(self, name, value=None):
        if name in self:
            return self[name]
        self[name] = value
        return value

    def getmany(self, names: List, return_first: bool = True):
        results = {}
        for name in names:
            if name in self:
                if return_first:
                    return self[name]
                results[name] = self[name]
        return results

    def __getattr__(self, name):
        if name in self:
            return self[name]
        return None

    def __setattr__(self, name, value):
        self[name] = value

    def __delattr__(self, name):
        if name in self:
            del self[name]
    
    @property
    def keys(self):
        return [k for k in self]

    def __getitem__(self, key):
        if isinstance(key, int):
            name = self.keys[key]
            return self[name]
        return self[key]


from __future__ import annotations

"""
ZStd Compression
"""

from .base import BaseCompression, logger
from typing import Optional

try:
    import zstd
    _zstd_available = True
except ImportError:
    _zstd_available = False


class ZstdCompression(BaseCompression):
    name: str = "zstd"
    compression_level: Optional[int] = 3


    def validate_compression_level(self):
        """
        Validates the compression level
        """
        assert self.compression_level <= 0 or \
            self.compression_level in range(23), \
            "Compression level must be between 0 and 22 or -131072 to -1"


    def check_deps(self):
        """
        Checks for dependencies
        """
        if _zstd_available is False:
            logger.error("zstd is not available. Please install `zstd` or `pyzstd` to use zstd compression")
            raise ImportError("zstd is not available. Please install `zstd` or `pyzstd` to use zstd compression")

    def compress(self, data: bytes, level: Optional[int] = None, **kwargs) -> bytes:
        """
        Compresses the data
        """
        if level is None: level = self.compression_level
        return zstd.compress(data, level)

    def decompress(self, data: bytes, **kwargs) -> bytes:
        """
        Decompresses the data
        """
        return zstd.decompress(data)
    
    




import os
import time
import socket
import pathlib
import tempfile
import contextlib

from typing import Union, Optional
from functools import lru_cache

__all__ = [
    'get_torch_device_name',
    'get_torch_device',
    'get_cpu_count',
    'is_readonly_dir',
    'get_host_name',
    'get_host_ip',
    'is_in_kubernetes',
    'get_k8s_namespace',
    'get_k8s_kubeconfig',
]

# Lazily get these libraries

@lru_cache()
def get_torch_device_name(mps_enabled: bool = False):
    with contextlib.suppress(ImportError):
        import torch
        if torch.cuda.is_available(): return 'cuda'
        if mps_enabled and torch.torch.backends.mps.is_available(): return 'mps'
    return 'cpu'

@lru_cache()
def get_torch_device(mps_enabled: bool = False):
    with contextlib.suppress(ImportError):
        import torch
        return torch.device(get_torch_device_name(mps_enabled))
    return 'cpu'

@lru_cache()
def get_cpu_count() -> int:
    """Get the available CPU count for this system.
    Takes the minimum value from the following locations:
    - Total system cpus available on the host.
    - CPU Affinity (if set)
    - Cgroups limit (if set)
    """
    import sys
    count = os.cpu_count()
    # Check CPU affinity if available
    with contextlib.suppress(Exception):
        import psutil
        affinity_count = len(psutil.Process().cpu_affinity())
        if affinity_count > 0:
            count = min(count, affinity_count)
    # Check cgroups if available
    if sys.platform == "linux":
        with contextlib.suppress(Exception):
            with open("/sys/fs/cgroup/cpu,cpuacct/cpu.cfs_quota_us") as f:
                quota = int(f.read())
            with open("/sys/fs/cgroup/cpu,cpuacct/cpu.cfs_period_us") as f:
                period = int(f.read())
            cgroups_count = quota // period
            if cgroups_count > 0:
                count = min(count, cgroups_count)

    return count


@lru_cache()
def is_readonly_dir(path: Union[str, pathlib.Path]) -> bool:
    """
    Check if a directory is read-only.
    """
    if isinstance(path, str):
        path = pathlib.Path(path)

    if not path.is_dir(): path = path.parent
    try:
        path.mkdir(parents = True, exist_ok = True)
        with tempfile.NamedTemporaryFile('w', dir = path.as_posix()) as f:
            f.write('test')
        return False
    except Exception as e:
        return True


_host_name = None
_host_ip = None

@lru_cache()
def get_host_name():
    """
    Wrapper to handle the case where the hostname is not yet available.
    """
    global _host_name
    if _host_name is None:
        while _host_name is None:
            with contextlib.suppress(Exception):
                _host_name = socket.gethostname()
            time.sleep(0.2)
    return _host_name


@lru_cache()
def get_host_ip():
    """
    Wrapper to handle the case where the hostname is not yet available.
    """
    global _host_ip
    if _host_ip is None:
        while _host_ip is None:
            with contextlib.suppress(Exception):
                _host_ip = socket.gethostbyname(get_host_name())
            time.sleep(0.2)
    return _host_ip


_is_in_kubernetes: bool = None
_k8s_namespace: str = None
_k8s_kubeconfig: str = None

@lru_cache()
def is_in_kubernetes() -> bool:
    """
    Check if we are running in a kubernetes cluster.
    """
    global _is_in_kubernetes
    if _is_in_kubernetes is None:
        _is_in_kubernetes = os.path.isdir('/var/run/secrets/kubernetes.io/')
    return _is_in_kubernetes


@lru_cache()
def get_k8s_namespace(
    env_var: str = 'NAMESPACE',
    default: str = 'default'
) -> str:
    """
    Get the kubernetes namespace.
    """
    global _k8s_namespace
    if _k8s_namespace is None:
        if is_in_kubernetes():
            with open('/var/run/secrets/kubernetes.io/serviceaccount/namespace') as f:
                _k8s_namespace = f.read().strip()
        else:
            _k8s_namespace = os.getenv(env_var, default)
    return _k8s_namespace

@lru_cache()
def get_k8s_kubeconfig(
    env_var: str = 'KUBECONFIG',
    default: str = None,
) -> str:
    """
    Get the kubernetes kubeconfig.
    """
    global _k8s_kubeconfig
    if _k8s_kubeconfig is None:
        if is_in_kubernetes():
            _k8s_kubeconfig = '/var/run/secrets/kubernetes.io/serviceaccount/token'
        else:
            _k8s_kubeconfig = os.getenv(env_var, default)
            if _k8s_kubeconfig is None:
                _k8s_kubeconfig = os.path.expanduser('~/.kube/config')
    return _k8s_kubeconfig


# Local K8s
_kubeconfig_dir: str = None

def get_local_kubeconfig_dir(
    env_var: str = 'KUBECONFIG_DIR',
    default: str = None,
) -> str:
    """
    Get the local kubernetes kubeconfig.
    """
    global _kubeconfig_dir
    if _kubeconfig_dir is None:
        _kubeconfig_dir = os.getenv(env_var, default) or os.path.expanduser('~/.kube')
    return _kubeconfig_dir

@lru_cache()
def get_local_kubeconfig(
    name: Optional[str] = None,
    set_as_envval: bool = False,
) -> str:
    """
    Get the local kubernetes kubeconfig
    - this assumes that it's not the default one.
    """
    if name is None: return get_k8s_kubeconfig()
    p = os.path.join(get_local_kubeconfig_dir(), name)
    for stem in ('', '-cluster', '-admin-config', '-config', '-kubeconfig'):
        for ext in ('', '.yaml', '.yml'):
            if os.path.isfile(p + stem + ext): 
                px = os.path.abspath(p + stem + ext)
                if set_as_envval:
                    os.environ['KUBECONFIG'] = px
                return px
    raise FileNotFoundError(f'Could not find kubeconfig file: {name} @ {p}')




from .nginx import Nginx, NginxConfig

from . import tfserving_pb2 as tfserving_config
from . import tfserving_api
from ._base import TFSModelVersion, TFSModelConfig
from .tfserving_pb2 import TFSConfig
from .tfserving_api import TFSModelEndpoint, TFServeModel

__all__ = [
    'Nginx',
    'NginxConfig',
    'tfserving_config',
    'tfserving_pb2',
    'tfserving_api',
    'TFSModelVersion', 
    'TFSModelConfig',
    'TFSConfig',
    'TFSModelEndpoint', 
    'TFServeModel'
]



from __future__ import annotations

from collections import defaultdict
from datetime import datetime
from typing import Sequence

from lazyops.libs.dbinit.base import TextClause, text

from lazyops.libs.dbinit.data_structures.grant_on import GrantOn, GrantStore
from lazyops.libs.dbinit.entities.cluster_entity import ClusterEntity
from lazyops.libs.dbinit.entities.entity import Entity
from lazyops.libs.dbinit.exceptions import EntityExistsError


class Role(ClusterEntity):
    """
    Represents a Postgres `Role <https://www.postgresql.org/docs/current/database-roles.html>`_.
    """

    def __init__(
        self,
        name: str,
        depends_on: Sequence[Entity] | None = None,
        check_if_exists: bool | None = None,
        superuser: bool | None = None,
        createdb: bool | None = None,
        createrole: bool | None = None,
        inherit: bool | None = None,
        login: bool | None = None,
        replication: bool | None = None,
        bypassrls: bool | None = None,
        connection_limit: int | None = None,
        password: str | None = None,
        encrypted: bool | None = None,
        valid_until: datetime | None = None,
        in_role: Sequence["Role"] | None = None,
        role: Sequence["Role"] | None = None,
        admin: Sequence["Role"] | None = None,
        grants: Sequence[GrantOn] | None = None,
    ):
        """
        All __init__ params correspond to CREATE ROLE arguments and options, see
        `official Postgres documentation <https://www.postgresql.org/docs/current/sql-createrole.html>`_.


        :param name: Unique name of the entity. Must be unique across the cluster.
        :param depends_on: Any entities that should be created before this one.
        :param check_if_exists: Flag to set existence check behavior. If `True`, will raise an exception during _safe_create if the entity already exists, and will raise an exception during _safe_drop if the entity does not exist.
        :param superuser: Flag to declare a role as a superuser. Postgres defaults to `False`.
        :param createdb: Flag for the ability to create databases. Postgres defaults to `False`.
        :param createrole: Flag for the ability to manage roles. Postgres defaults to `False`.
        :param inherit: Flag to specify that the role will inherit privileges of roles it is a member of. Postgres defaults to `True`.
        :param login: Flag for the ability to login (aka be a user). Postgres defaults to `False`.
        :param replication: Flag to specify a replication role. Postgres defaults to `False`.
        :param bypassrls: Flag to indicate the role should bypass RLS policies. Postgres defaults to `False`.
        :param connection_limit: Number of concurrent connections the role can make. Postgres defaults to -1, which means no limit.
        :param password: Set the role's password. Postgres defaults to a null password, which means password authentication always fails.
        :param encrypted: Keyword has no effect but is present for backwards compatability. Consider removing.
        :param valid_until: Sets a date and time after which the role's password is no longer valid. Postgres defaults to no limit.
        :param in_role: Sequence of :class:`lazyops.libs.dbinit.entities.Role` that this role belongs to.
        :param role: Sequence of :class:`lazyops.libs.dbinit.entities.Role` that belong to this role.
        :param admin: Like the `role` param above, but gives the roles the right to grant membership to this role to other roles.
        :param grants: Sequence of :class:`lazyops.libs.dbinit.data_structures.GrantOn` to specify privileges this role has in relation to other entities.
        """
        self.superuser = superuser
        self.createdb = createdb
        self.createrole = createrole
        self.inherit = inherit
        self.login = login
        self.replication = replication
        self.bypassrls = bypassrls
        self.connection_limit = connection_limit
        self.password = password
        self.encrypted = encrypted
        self.valid_until = valid_until
        self.in_role = in_role
        self.role = role
        self.admin = admin
        self.grants: GrantStore = defaultdict(set)
        if grants:
            self.grant(grants=grants)
        super().__init__(name=name, depends_on=depends_on, check_if_exists=check_if_exists)

    def _create_statements(self) -> Sequence[TextClause]:
        # statement = f"CREATE ROLE {self.name}"
        statement = f"CREATE ROLE {self.encoded_name}"
        props = self._get_passed_args()

        for k, v in props.items():
            if k == "password" and isinstance(v, str):
                if "encrypted" in props:
                    statement = f"{statement} ENCRYPTED"
                statement = f"{statement} PASSWORD '{v}'"
                if "valid_until" in props:
                    timestamp = props.get("valid_until")
                    statement = f"{statement} VALID UNTIL '{timestamp}'"
            
            elif isinstance(v, bool) and k != "encrypted":
                flag = k.upper()
                if not v:
                    flag = f"NO{flag}"
                statement = f"{statement} {flag}"
            
            elif k == "connection_limit" and isinstance(v, int):
                statement = f"{statement} CONNECTION LIMIT {v}"
            
            elif k == "grants":
                # ignore grants argument, it is handled by the _grant command
                pass

            elif isinstance(v, list) and len(v) > 0:
                [first, *roles] = v
                option = k.upper().replace("_", " ")
                roles = [first] + roles  # type: ignore
                formatted_roles = ", ".join([r.name for r in roles])
                statement = f"{statement} {option} {formatted_roles}"

        # TODO binding isn't working, switching to simple quotes for now
        return [text(statement)]

    def _exists_statement(self) -> TextClause:
        # return text("SELECT EXISTS(SELECT 1 FROM pg_authid WHERE rolname=:role)").bindparams(role=self.encoded_name)
        return text("SELECT EXISTS(SELECT 1 FROM pg_authid WHERE rolname=:role)").bindparams(role=self.name)

    def _drop_statements(self) -> Sequence[TextClause]:
        return [text(f"DROP ROLE {self.encoded_name}")]
        # return [text(f"DROP ROLE {self.name}")]

    def grant(self, grants: Sequence[GrantOn]) -> None:
        """
        Sets the privileges to grant on this role to the specified entities.
        :param grants: A Sequence of :class:`lazyops.libs.dbinit.data_structures.GrantOn` to store.
        """
        for grant in grants:
            for target in grant.on:
                self.grants[target] = self.grants[target].union(set(grant.privileges))

    def _safe_grant(self) -> None:
        """
        Performs an existence check before executing grant statements in the cluster.
        """
        if self.grants:
            if not self._exists():
                raise EntityExistsError(
                    f"There is no {self.__class__.__name__} with the "
                    f"name {self.name}. The {self.__class__.__name__} "
                    f"must exist to grant privileges."
                )
            for target, privileges in self.grants.items():
                target._safe_grant(grantee=self, privileges=privileges)

    def _grants_exist(self) -> bool:
        """
        Checks to see if all in-code declared grants exist in the cluster.
        """
        if not self.grants:
            # if there aren't any grants, return True because technically all specified grants are present
            return True
        if not self._exists():
            raise EntityExistsError(
                f"There is no {self.__class__.__name__} with the "
                f"name {self.name}. The {self.__class__.__name__} "
                f"must exist to alter privileges."
            )
        else:
            return all(
                target._grants_exist(grantee=self, privileges=privileges)
                for target, privileges in self.grants.items()
            )

    def _safe_revoke(self) -> None:
        """
        Performs an existence check before executing revoke statements in the cluster.
        """
        if self.grants:
            if not self._exists():
                raise EntityExistsError(
                    f"There is no {self.__class__.__name__} with the "
                    f"name {self.name}. The {self.__class__.__name__} "
                    f"must exist to revoke privileges."
                )
            for target, privileges in self.grants.items():
                target._safe_revoke(grantee=self, privileges=privileges)

    def __repr__(self):
        return f"<Role (name={self.name}, superuser={self.superuser}, createdb={self.createdb}, createrole={self.createrole}, inherit={self.inherit}, login={self.login}, replication={self.replication}, bypassrls={self.bypassrls}, connection_limit={self.connection_limit}, password={'*****' if self.password else None}, encrypted={self.encrypted}, valid_until={self.valid_until}, in_role={self.in_role}, role={self.role}, admin={self.admin}, grants={self.grants})>"


from .logs import logger
from .logs import change_logger_level, add_api_log_filters

from .lazylib import LazyLib
from .imports import (
    resolve_missing,
    resolve_missing_custom,
    require_missing_wrapper,
)
from .serialization import (
    object_serializer, object_deserializer,
    ObjectEncoder, ObjectDecoder,
    Json,
)
from .lazy import (
    import_string, 
    lazy_import, 
    is_coro_func,
    import_function,
    validate_callable,
    fetch_property,
    extract_base_model_kws,

)
from .helpers import (
    timer,
    timed,
    merge_dicts,
    
    exponential_backoff,
    retryable,
    build_batches,
    get_batches_from_generator,
    split_into_batches,
    split_into_batches_with_index,

    create_timestamp,
    create_unique_id,
    create_secret,

    suppress,
    timed_cache,
    fail_after,
    create_background_task,
)   

from .dates import (
    convert_date,
    parse_datetime,
    parse_datetime_from_timestamp,
    get_current_datetime,
    is_expired_datetime,
)
from .times import (
    Timer
)





from lazyops.imports._k8s import (
    resolve_k8s,
)

resolve_k8s(
    is_async=True,
    is_sync=True,
    is_operator=True,
    required=True,
)

import kubernetes.client as SyncClient
import kubernetes.config as SyncConfig
import kubernetes.stream.ws_client as SyncStream
import kubernetes.stream.stream as SyncStreamFunc

import kubernetes.watch as SyncWatch
import kubernetes.utils as SyncUtils
import kubernetes.client.models as SyncType

import kubernetes_asyncio.client as AsyncClient
import kubernetes_asyncio.config as AsyncConfig
import kubernetes_asyncio.stream as AsyncStream
import kubernetes_asyncio.watch as AsyncWatch
import kubernetes_asyncio.utils as AsyncUtils
import kubernetes_asyncio.client.models as AsyncType

AsyncWSClient = AsyncStream.WsApiClient



# from lazyops.imports._k8s import (
#     AsyncClient,
#     AsyncConfig,
#     AsyncStream,
#     AsyncWatch,
#     AsyncUtils,
#     AsyncType,
#     AsyncWSClient,

#     SyncClient,
#     SyncConfig,
#     SyncStream,
#     SyncStreamFunc,

#     SyncWatch,
#     SyncUtils,
#     SyncType,

# )


from __future__ import annotations

import random
import asyncio
import logging
import contextlib
from enum import Enum
from lazyops.libs.kops.base import *
from lazyops.libs.kops.config import KOpsSettings
from lazyops.libs.kops.utils import cached, DillSerializer, SignalHandler
from lazyops.libs.kops._kopf import kopf
from lazyops.types import lazyproperty
from lazyops.utils import logger


from typing import List, Dict, Union, Any, Optional, Callable, TYPE_CHECKING

import lazyops.libs.kops.types as t
import lazyops.libs.kops.atypes as at

if TYPE_CHECKING:
    with contextlib.suppress(ImportError):
        import requests
        import aiohttpx


class EventType(str, Enum):
    startup = 'startup'
    shutdown = 'shutdown'


class KOpsContext:
    def __init__(
        self,
        settings: KOpsSettings,
        ctx: Optional[str] = None,
        name: Optional[str] = None,
        config_file: Optional[str] = None,

    ):
        self.settings = settings
        self.ctx = ctx
        self.name = name
        self.config_file = config_file
        self.ainitialized: bool = False
        self.initialized: bool = False

    async def aset_k8_config(self):
        if self.ainitialized: return
        if self.settings.in_k8s:
            logger.info('Loading in-cluster config')
            AsyncConfig.load_incluster_config()
        else:
            config = self.settings.get_kconfig_path(self.ctx or self.name)
            if config: config = config.as_posix()
            logger.info(f'Loading kubeconfig from {config}')
            await AsyncConfig.load_kube_config(config_file = config, context = self.ctx or self.name or self.settings.kubeconfig_context)
        logger.info('Initialized K8s Client')
        self.ainitialized = True
    
    def set_k8_config(self):
        if self.initialized: return
        if self.settings.in_k8s:
            logger.info('Loading in-cluster config')
            SyncConfig.load_incluster_config()
        else:
            config = self.settings.get_kconfig_path(self.ctx or self.name)
            if config: config = config.as_posix()
            logger.info(f'Loading kubeconfig from {config}')
            SyncConfig.load_kube_config(config_file = config, context = self.ctx or self.name or self.settings.kubeconfig_context)
        logger.info('Initialized K8s Client')
        self.initialized = True

    @lazyproperty
    def client(self) -> 'SyncClient.ApiClient':
        """
        Primary Sync Client
        """
        return SyncClient.ApiClient(pool_threads=4)

    @lazyproperty
    def aclient(self) -> 'AsyncClient.ApiClient':
        """
        Primary Async Client
        """
        return AsyncClient.ApiClient(pool_threads=4)
    
    @lazyproperty
    def wsclient(self) -> 'SyncStream.WSClient':
        """
        Primary Websocket Client
        """
        return SyncStream.WSClient(pool_threads=4)
    
    @lazyproperty
    def awsclient(self) -> 'AsyncStream.WsApiClient':
        """
        Primary Async Websocket Client
        """
        return AsyncStream.WsApiClient()
    
    @lazyproperty
    def core_v1(self) -> 'SyncClient.CoreV1Api':
        """
        - StatefulSets
        - Deployments
        - DaemonSets
        - ReplicaSets
        - Secrets
        - Pods
        - Nodes
        """
        return SyncClient.CoreV1Api(self.client)
    
    @lazyproperty
    def core_v1_ws(self) -> 'SyncClient.CoreV1Api':
        """
        Websocket Client

        - StatefulSets
        - Deployments
        - DaemonSets
        - ReplicaSets
        - Secrets
        - Pods
        - Nodes
        """
        return SyncClient.CoreV1Api(self.wsclient)
    

    @lazyproperty
    def apps_v1(self) -> 'SyncClient.AppsV1Api':
        """
        - StatefulSets
        - Deployments
        - DaemonSets
        - ReplicaSets
        """
        return SyncClient.AppsV1Api(self.client)

    @lazyproperty
    def networking_v1(self) -> 'SyncClient.NetworkingV1Api':
        """
        - Ingress
        """
        return SyncClient.NetworkingV1Api(self.client)
    

    @lazyproperty
    def crds(self) -> 'SyncClient.ApiextensionsV1Api':
        return SyncClient.ApiextensionsV1Api(self.client)

    
    @lazyproperty
    def customobjs(self) -> 'SyncClient.CustomObjectsApi':
        return SyncClient.CustomObjectsApi(self.client)

    """
    Async Properties
    """

    @lazyproperty
    def acore_v1(self) -> 'AsyncClient.CoreV1Api':
        """
        - StatefulSets
        - Deployments
        - DaemonSets
        - ReplicaSets
        - Secrets
        - Pods
        - Nodes
        """
        return AsyncClient.CoreV1Api(self.aclient)
    

    @lazyproperty
    def acore_v1_ws(self) -> 'AsyncClient.CoreV1Api':
        """
        Websocket Client

        - StatefulSets
        - Deployments
        - DaemonSets
        - ReplicaSets
        - Secrets
        - Pods
        - Nodes
        """
        return AsyncClient.CoreV1Api(self.awsclient)
    

    @lazyproperty
    def aapps_v1(self) -> 'AsyncClient.AppsV1Api':
        """
        - StatefulSets
        - Deployments
        - DaemonSets
        - ReplicaSets
        """
        return AsyncClient.AppsV1Api(self.aclient)

    @lazyproperty
    def anetworking_v1(self) -> 'AsyncClient.NetworkingV1Api':
        """
        - Ingress
        """
        return AsyncClient.NetworkingV1Api(self.aclient)
    

    @lazyproperty
    def acrds(self) -> 'AsyncClient.ApiextensionsV1Api':
        return AsyncClient.ApiextensionsV1Api(self.aclient)

    
    @lazyproperty
    def acustomobjs(self) -> 'AsyncClient.CustomObjectsApi':
        return AsyncClient.CustomObjectsApi(self.aclient)
    


    """
    Sync Resource Level
    """
    @lazyproperty
    def config_maps(self) -> 'SyncClient.CoreV1Api':
        return self.core_v1
    
    @lazyproperty
    def secrets(self) -> 'SyncClient.CoreV1Api':
        return self.core_v1
    
    @lazyproperty
    def pods(self) -> 'SyncClient.CoreV1Api':
        return self.core_v1
    
    @lazyproperty
    def nodes(self) -> 'SyncClient.CoreV1Api':
        return self.core_v1
    
    @lazyproperty
    def services(self) -> 'SyncClient.CoreV1Api':
        return self.core_v1
    
    @lazyproperty
    def ingresses(self) -> 'SyncClient.NetworkingV1Api':
        return self.networking_v1
    
    @lazyproperty
    def stateful_sets(self) -> 'SyncClient.AppsV1Api':
        return self.apps_v1
    
    @lazyproperty
    def deployments(self) -> 'SyncClient.AppsV1Api':
        return self.apps_v1
    
    @lazyproperty
    def daemon_sets(self) -> 'SyncClient.AppsV1Api':
        return self.apps_v1
    
    @lazyproperty
    def replica_sets(self) -> 'SyncClient.AppsV1Api':
        return self.apps_v1
    
    @lazyproperty
    def customresourcedefinitions(self) -> 'SyncClient.ApiextensionsV1Api':
        return self.crds
    
    @lazyproperty
    def customobjects(self) -> 'SyncClient.CustomObjectsApi':
        return self.customobjs
    
    @lazyproperty
    def persistent_volumes(self) -> 'SyncClient.CoreV1Api':
        return self.core_v1
    
    @lazyproperty
    def persistent_volume_claims(self) -> 'SyncClient.CoreV1Api':
        return self.core_v1
    
    """
    Async Resource Level
    """

    @lazyproperty
    def aconfig_maps(self) -> 'AsyncClient.CoreV1Api':
        return self.acore_v1
    
    @lazyproperty
    def asecrets(self) -> 'AsyncClient.CoreV1Api':
        return self.acore_v1
    
    @lazyproperty
    def apods(self) -> 'AsyncClient.CoreV1Api':
        return self.acore_v1
    
    @lazyproperty
    def anodes(self) -> 'AsyncClient.CoreV1Api':
        return self.acore_v1
    
    @lazyproperty
    def aservices(self) -> 'AsyncClient.CoreV1Api':
        return self.acore_v1
    
    @lazyproperty
    def aingresses(self) -> 'AsyncClient.NetworkingV1Api':
        return self.anetworking_v1
    
    @lazyproperty
    def astateful_sets(self) -> 'AsyncClient.AppsV1Api':
        return self.aapps_v1
    
    @lazyproperty
    def adeployments(self) -> 'AsyncClient.AppsV1Api':
        return self.aapps_v1
    
    @lazyproperty
    def adaemon_sets(self) -> 'AsyncClient.AppsV1Api':
        return self.aapps_v1
    
    @lazyproperty
    def areplica_sets(self) -> 'AsyncClient.AppsV1Api':
        return self.aapps_v1
    
    @lazyproperty
    def acustomresourcedefinitions(self) -> 'AsyncClient.ApiextensionsV1Api':
        return self.acrds
    
    @lazyproperty
    def acustomobjects(self) -> 'AsyncClient.CustomObjectsApi':
        return self.acustomobjs
    
    @lazyproperty
    def apersistent_volumes(self) -> 'AsyncClient.CoreV1Api':
        return self.acore_v1
    
    @lazyproperty
    def apersistent_volume_claims(self) -> 'AsyncClient.CoreV1Api':
        return self.acore_v1
    
    @property
    def auth_headers(self) -> Dict[str, str]:
        if not self.initialized and not self.ainitialized: return None
        auth = self.aclient.configuration.auth_settings() if self.ainitialized \
            else self.client.configuration.auth_settings()
        return {auth['BearerToken']['key']: auth['BearerToken']['value']}

    @property
    def request_headers(self) -> Dict[str, str]:
        if not self.initialized and not self.ainitialized: return None
        headers = self.aclient.default_headers if self.ainitialized \
            else self.client.default_headers
        # logger.info(f"Request Headers: {headers}")
        return {
            'Accept': 'application/json;as=Table;v=v1;g=meta.k8s.io,application/json;as=Table;v=v1beta1;g=meta.k8s.io,application/json',
            'Connection': 'keep-alive',
            # 'Keep-alive': 'max=10, timeout=1200',
            **self.auth_headers,
            **headers
        }
    
    @property
    def cluster_url(self) -> str:
        if self.initialized or self.ainitialized:
            return self.aclient.configuration.host if self.ainitialized \
                    else self.client.configuration.host
        else: return None

    @property
    def ssl_ca_cert(self) -> str:
        if self.initialized or self.ainitialized:
            return self.aclient.configuration.ssl_ca_cert if self.ainitialized \
                    else self.client.configuration.ssl_ca_cert
        else: return None

    @lazyproperty
    def http_client(self) -> Union['aiohttpx.Client', 'requests.Session']:
        """
        Returns the http client
        """
        with contextlib.suppress(ImportError):
            import aiohttpx
            return aiohttpx.Client(
                base_url = self.cluster_url,
                headers = self.request_headers, 
                verify = self.ssl_ca_cert,
            )
        with contextlib.suppress(ImportError):
            import requests
            sess = requests.Session()
            sess.headers.update(self.request_headers)
            sess.verify = self.ssl_ca_cert
            return sess
        raise ImportError('No http client found. Please install aiohttpx or requests')


    @staticmethod
    def to_singular(resource: str) -> str:
        if resource.endswith('es'):
            return resource[:-2]
        elif resource.endswith('s'):
            return resource[:-1]
        return resource


    """
    API Methods
    """


    def get(self, resource: str, name: str, namespace: str = None, **kwargs) -> 'SyncClient.V1ObjectMeta':
        api = getattr(self, resource)
        singular = self.to_singular(resource)
        return getattr(api, f'read_namespaced_{singular}')(name, namespace = namespace, **kwargs)
    
    def list(self, resource: str, namespace: str = None, **kwargs) -> 'SyncClient.V1ObjectMeta':
        api = getattr(self, resource)
        singular = self.to_singular(resource)
        if namespace:
            return getattr(api, f'list_namespaced_{singular}')(namespace = namespace, **kwargs)
        return getattr(api, f'list_{singular}_for_all_namespaces')(**kwargs)
    
    def create(self, resource: str, namespace: str = None, **kwargs) -> 'SyncClient.V1ObjectMeta':
        api = getattr(self, resource)
        singular = self.to_singular(resource)
        return getattr(api, f'create_namespaced_{singular}')(namespace = namespace, **kwargs)
    
    def update(self, resource: str, name: str, namespace: str = None, **kwargs) -> 'SyncClient.V1ObjectMeta':
        api = getattr(self, resource)
        singular = self.to_singular(resource)
        return getattr(api, f'patch_namespaced_{singular}')(name, namespace = namespace, **kwargs)
    
    def delete(self, resource: str, name: str, namespace: str = None, **kwargs) -> 'SyncClient.V1ObjectMeta':
        api = getattr(self, resource)
        singular = self.to_singular(resource)
        return getattr(api, f'delete_namespaced_{singular}')(name, namespace = namespace, **kwargs)
    
    def patch(self, resource: str, name: str, namespace: str = None, **kwargs) -> 'SyncClient.V1ObjectMeta':
        api = getattr(self, resource)
        singular = self.to_singular(resource)
        return getattr(api, f'patch_namespaced_{singular}')(name, namespace = namespace, **kwargs)


    async def aget(self, resource: str, name: str, namespace: str = None, **kwargs) -> 'AsyncClient.V1ObjectMeta':
        aresource = resource if resource.startswith('a') else f'a{resource}'
        api = getattr(self, aresource)
        singular = self.to_singular(resource)
        return await getattr(api, f'read_namespaced_{singular}')(name, namespace = namespace, **kwargs)
    
    async def alist(self, resource: str, namespace: str = None, **kwargs) -> 'AsyncClient.V1ObjectMeta':
        aresource = resource if resource.startswith('a') else f'a{resource}'
        api = getattr(self, aresource)
        singular = self.to_singular(resource)
        if namespace:
            return await getattr(api, f'list_namespaced_{singular}')(namespace = namespace, **kwargs)
        return await getattr(api, f'list_{singular}_for_all_namespaces')(**kwargs)
    
    async def acreate(self, resource: str, namespace: str = None, **kwargs) -> 'AsyncClient.V1ObjectMeta':
        aresource = resource if resource.startswith('a') else f'a{resource}'
        api = getattr(self, aresource)
        singular = self.to_singular(resource)
        return await getattr(api, f'create_namespaced_{singular}')(namespace = namespace, **kwargs)
    
    async def aupdate(self, resource: str, name: str, namespace: str = None, **kwargs) -> 'AsyncClient.V1ObjectMeta':
        aresource = resource if resource.startswith('a') else f'a{resource}'
        api = getattr(self, aresource)
        singular = self.to_singular(resource)
        return await getattr(api, f'patch_namespaced_{singular}')(name, namespace = namespace, **kwargs)
    
    async def adelete(self, resource: str, name: str, namespace: str = None, **kwargs) -> 'AsyncClient.V1ObjectMeta':
        aresource = resource if resource.startswith('a') else f'a{resource}'
        api = getattr(self, aresource)
        singular = self.to_singular(resource)
        return await getattr(api, f'delete_namespaced_{singular}')(name, namespace = namespace, **kwargs)
    
    async def apatch(self, resource: str, name: str, namespace: str = None, **kwargs) -> 'AsyncClient.V1ObjectMeta':
        aresource = resource if resource.startswith('a') else f'a{resource}'
        api = getattr(self, aresource)
        singular = self.to_singular(resource)
        return await getattr(api, f'patch_namespaced_{singular}')(name, namespace = namespace, **kwargs)




class KOpsClientMeta(type):
    """
    This is the metaclass for the KOpsClient class.
    """

    ctx: Optional[str] = None
    _settings: Optional[KOpsSettings] = None
    _session: Optional[KOpsContext] = None
    _sessions: Dict[str, KOpsContext] = {}
    _startup_functions: List[Callable] = []
    _shutdown_functions: List[Callable] = []

    @property
    def settings(cls) -> KOpsSettings:
        if not cls._settings:
            cls._settings = KOpsSettings()
        return cls._settings
    
    def add_session(
        cls,
        name: Optional[str] = None, 
        config_file: Optional[str] = None, 
        ctx: Optional[Any] = None, 
        set_as_current: Optional[bool] = True, 
        **kwargs
    ):
        """
        Adds a session to the client.
        """
        name = name or 'default'
        if name not in cls._sessions:
            cls._sessions[name] = KOpsContext(
                settings = cls.settings,
                config_file = config_file,
                ctx = ctx,
                **kwargs
            )
            logger.info(f'Session created: {name}')
        if set_as_current:
            cls.ctx = name
            cls._session = cls._sessions[name]
            logger.info(f'Session set: {name}')
    
    def set_session(
        cls,
        name: Optional[str] = None,
    ):
        """
        Sets the current session.
        """
        name = name or 'default'
        if name in cls._sessions:
            cls.ctx = name
            cls._session = cls._sessions[name]
            logger.info(f'Session set: {name}')

    @property
    def session(cls) -> KOpsContext:
        if not cls._session:
            cls._session = KOpsContext(
                settings = cls.settings,
                ctx = cls.ctx
            )
            if cls.ctx is None: cls.ctx = 'default'
            cls._sessions[cls.ctx] = cls._session
            logger.info(f'Session set: {cls.ctx}')
        return cls._session

    @property
    def client(cls) -> 'SyncClient.ApiClient':
        """
        Returns the kubernetes client.
        """
        return cls.session.client

    @property
    def aclient(cls) -> 'AsyncClient.ApiClient':
        """
        Returns the async kubernetes client.
        """
        return cls.session.aclient
    
    """
    HTTP Client Properties
    """

    @property
    def auth_headers(cls) -> Dict[str, str]:
        """
        Returns the authentication headers.
        """
        return cls.session.auth_headers
    
    @property
    def request_headers(cls) -> Dict[str, str]:
        """
        Returns the request headers.
        """
        return cls.session.request_headers
    
    @property
    def cluster_url(cls) -> str:
        """
        Returns the cluster url.
        """
        return cls.session.cluster_url
    
    @property
    def ssl_ca_cert(cls) -> str:
        """
        Returns the ssl ca cert.
        """
        return cls.session.ssl_ca_cert
    
    @property
    def http_client(cls) -> Union['aiohttpx.Client', 'requests.Session']:
        """
        Returns the http client.
        """
        return cls.session.http_client

    @property
    def wsclient(cls) -> 'SyncStream.WSClient':
        """
        Returns the websocket kubernetes client.
        """
        return cls.session.wsclient
    
    @property
    def awsclient(cls) -> 'AsyncStream.WsApiClient':
        """
        Returns the async websocket kubernetes client.
        """
        return cls.session.awsclient
    
    
    @property
    def apps_v1(cls) -> 'SyncClient.AppsV1Api':
        """
        Returns the apps_v1 api.
        """
        return cls.session.apps_v1

    @property
    def core_v1(cls) -> 'SyncClient.CoreV1Api':
        """
        Returns the core_v1 api.
        """
        return cls.session.core_v1
    
    @property
    def core_v1_ws(cls) -> 'SyncClient.CoreV1Api':
        """
        Websocket Client:
        - StatefulSets
        - Deployments
        - DaemonSets
        - ReplicaSets
        - Secrets
        - Pods
        - Nodes
        """
        return cls.session.core_v1_ws
    
    @property
    def crds(cls) -> 'SyncClient.CustomObjectsApi':
        """
        Returns the crds_v1 api.
        """
        return cls.session.crds

    
    @property
    def customobjs(cls) -> 'SyncClient.CustomObjectsApi':
        return cls.session.customobjs
    

    """
    Async Properties
    """

    @property
    def acore_v1(cls) -> 'AsyncClient.CoreV1Api':
        """
        - StatefulSets
        - Deployments
        - DaemonSets
        - ReplicaSets
        - Secrets
        - Pods
        - Nodes
        """
        return cls.session.acore_v1
    
    @property
    def acore_v1_ws(cls) -> 'AsyncClient.CoreV1Api':
        """
        Websocket Client:
        - StatefulSets
        - Deployments
        - DaemonSets
        - ReplicaSets
        - Secrets
        - Pods
        - Nodes
        """
        return cls.session.acore_v1_ws
    

    @property
    def aapps_v1(cls) -> 'AsyncClient.AppsV1Api':
        """
        - StatefulSets
        - Deployments
        - DaemonSets
        - ReplicaSets
        """
        return cls.session.aapps_v1

    @property
    def anetworking_v1(cls) -> 'AsyncClient.NetworkingV1Api':
        """
        - Ingress
        """
        return cls.session.anetworking_v1
    

    @property
    def acrds(cls) -> 'AsyncClient.ApiextensionsV1Api':
        return cls.session.acrds

    
    @property
    def acustomobjs(cls) -> 'AsyncClient.CustomObjectsApi':
        return cls.session.acustomobjs
    


    """
    Sync Resource Level
    """
    @property
    def config_maps(cls) -> 'SyncClient.CoreV1Api':
        return cls.session.config_maps
    
    @property
    def secrets(cls) -> 'SyncClient.CoreV1Api':
        return cls.session.secrets
    
    @property
    def pods(cls) -> 'SyncClient.CoreV1Api':
        return cls.session.pods
    
    @property
    def nodes(cls) -> 'SyncClient.CoreV1Api':
        return cls.session.nodes
    
    @property
    def services(cls) -> 'SyncClient.CoreV1Api':
        return cls.session.services
    
    @property
    def ingresses(cls) -> 'SyncClient.NetworkingV1Api':
        return cls.session.ingresses
    
    @property
    def stateful_sets(cls) -> 'SyncClient.AppsV1Api':
        return  cls.session.stateful_sets
    
    @property
    def deployments(cls) -> 'SyncClient.AppsV1Api':
        return cls.session.deployments
    
    @property
    def daemon_sets(cls) -> 'SyncClient.AppsV1Api':
        return cls.session.daemon_sets
    
    @property
    def replica_sets(cls) -> 'SyncClient.AppsV1Api':
        return cls.session.replica_sets
    
    @property
    def customresourcedefinitions(cls) -> 'SyncClient.ApiextensionsV1Api':
        return cls.session.customresourcedefinitions
    
    @property
    def customobjects(cls) -> 'SyncClient.CustomObjectsApi':
        return cls.session.customobjects
    
    @property
    def persistent_volumes(cls) -> 'SyncClient.CoreV1Api':
        return cls.session.persistent_volumes
    
    @property
    def persistent_volume_claims(cls) -> 'SyncClient.CoreV1Api':
        return cls.session.persistent_volume_claims
    
    """
    Async Resource Level
    """

    @property
    def aconfig_maps(cls) -> 'AsyncClient.CoreV1Api':
        return cls.session.aconfig_maps
    
    @property
    def asecrets(cls) -> 'AsyncClient.CoreV1Api':
        return cls.session.asecrets
    
    @property
    def apods(cls) -> 'AsyncClient.CoreV1Api':
        return cls.session.apods
    
    @property
    def anodes(cls) -> 'AsyncClient.CoreV1Api':
        return cls.session.anodes
    
    @property
    def aservices(cls) -> 'AsyncClient.CoreV1Api':
        return cls.session.aservices
    
    @property
    def aingresses(cls) -> 'AsyncClient.NetworkingV1Api':
        return cls.session.aingresses
    
    @property
    def astateful_sets(cls) -> 'AsyncClient.AppsV1Api':
        return cls.session.astateful_sets
    
    @property
    def adeployments(cls) -> 'AsyncClient.AppsV1Api':
        return cls.session.adeployments
    
    @property
    def adaemon_sets(cls) -> 'AsyncClient.AppsV1Api':
        return cls.session.adaemon_sets
    
    @property
    def areplica_sets(cls) -> 'AsyncClient.AppsV1Api':
        return cls.session.areplica_sets
    
    @property
    def acustomresourcedefinitions(cls) -> 'AsyncClient.ApiextensionsV1Api':
        return cls.session.acustomresourcedefinitions
    
    @property
    def acustomobjects(cls) -> 'AsyncClient.CustomObjectsApi':
        return cls.session.acustomobjects
    
    @property
    def apersistent_volumes(cls) -> 'AsyncClient.CoreV1Api':
        return cls.session.apersistent_volumes
    
    @property
    def apersistent_volume_claims(cls) -> 'AsyncClient.CoreV1Api':
        return cls.session.apersistent_volume_claims

    async def aconfigure(cls):
        """
        Sets the kubernetes config for the current context.
        """
        await cls.session.aset_k8_config()

    async def aset_k8_config(cls):
        """
        Sets the kubernetes config for the current context.
        """
        await cls.session.aset_k8_config()
    
    def configure(cls):
        """
        Sets the kubernetes config for the current context.
        """
        cls.session.set_k8_config()

    def set_k8_config(cls):
        """
        Sets the kubernetes config for the current context.
        """
        cls.session.set_k8_config()

    
    def get(cls, resource: str, name: str, namespace: str = None, **kwargs) -> 'SyncClient.V1ObjectMeta':
        """
        Gets a resource.
        """
        return cls.session.get(resource, name, namespace, **kwargs)
    
    
    def list(cls, resource: str, namespace: str = None, **kwargs) -> 'SyncClient.V1ObjectMeta':
        """
        Lists a resource.
        """
        return cls.session.list(resource, namespace, **kwargs)
    
    def create(cls, resource: str, namespace: str = None, **kwargs) -> 'SyncClient.V1ObjectMeta':
        """
        Creates a resource.
        """
        return cls.session.create(resource, namespace, **kwargs)
        
    def update(cls, resource: str, name: str, namespace: str = None, **kwargs) -> 'SyncClient.V1ObjectMeta':
        """
        Updates a resource.
        """
        return cls.session.update(resource, name, namespace, **kwargs)
    
    def delete(cls, resource: str, name: str, namespace: str = None, **kwargs) -> 'SyncClient.V1ObjectMeta':
        """
        Deletes a resource.
        """
        return cls.session.delete(resource, name, namespace, **kwargs)
    
    def patch(cls, resource: str, name: str, namespace: str = None, **kwargs) -> 'SyncClient.V1ObjectMeta':
        """
        Patches a resource.
        """
        return cls.session.patch(resource, name, namespace, **kwargs)


    async def aget(cls, resource: str, name: str, namespace: str = None, **kwargs) -> 'AsyncClient.V1ObjectMeta':
        """
        Gets a resource.
        """
        return await cls.session.aget(resource, name, namespace, **kwargs)

    
    async def alist(cls, resource: str, namespace: str = None, **kwargs) -> 'AsyncClient.V1ObjectMeta':
        """
        Lists a resource.
        """
        return await cls.session.alist(resource, namespace, **kwargs)
    
    async def acreate(cls, resource: str, namespace: str = None, **kwargs) -> 'AsyncClient.V1ObjectMeta':
        """
        Creates a resource.
        """
        return await cls.session.acreate(resource, namespace, **kwargs)
    
    async def aupdate(cls, resource: str, name: str, namespace: str = None, **kwargs) -> 'AsyncClient.V1ObjectMeta':
        """
        Updates a resource.
        """
        return await cls.session.aupdate(resource, name, namespace, **kwargs)
    
    async def adelete(cls, resource: str, name: str, namespace: str = None, **kwargs) -> 'AsyncClient.V1ObjectMeta':
        """
        Deletes a resource.
        """
        return await cls.session.adelete(resource, name, namespace, **kwargs)
    
    async def apatch(cls, resource: str, name: str, namespace: str = None, **kwargs) -> 'AsyncClient.V1ObjectMeta':
        """
        Patches a resource.
        """
        return await cls.session.apatch(resource, name, namespace, **kwargs)

    """
    KOpf Methods
    """

    def add_function(cls, function: Callable, event: EventType = EventType.startup):
        if event == EventType.startup:
            cls._startup_functions.append(function)
        elif event == EventType.shutdown:
            cls._shutdown_functions.append(function)
    
    async def run_startup_functions(cls, **kwargs):
        # Start Signal Watcher
        asyncio.create_task(SignalHandler.monitor(cls._shutdown_functions))

        if not cls._startup_functions: return
        for func in cls._startup_functions:
            await func(**kwargs)

    async def run_shutdown_functions(cls, **kwargs):
        if not cls._shutdown_functions: return
        for func in cls._shutdown_functions:
            await func(**kwargs)
    

    # Register Startup Functions
    def configure_kopf(
        cls, 
        _logger: logging.Logger = None,
        _settings: Optional[kopf.OperatorSettings] = None,
        enable_event_logging: Optional[bool] = None,
        event_logging_level: Optional[str] = None,
        finalizer: Optional[str] = None,
        storage_prefix: Optional[str] = None,
        persistent_key: Optional[str] = None,
        error_delays: Optional[List[int]] = None,
        startup_functions: Optional[List[Callable]] = None,
        shutdown_functions: Optional[List[Callable]] = None,
        kopf_name: Optional[str] = None,
        app_name: Optional[str] = None,
        multi_pods: Optional[bool] = True,
        **kwargs
    ):
        """
        Registers the startup function for configuring kopf.

        Parameters
        """ 
        enable_event_logging = enable_event_logging if enable_event_logging is not None \
            else cls.settings.kopf_enable_event_logging
        event_logging_level = event_logging_level if event_logging_level is not None \
            else cls.settings.kopf_event_logging_level
        finalizer = finalizer if finalizer is not None \
            else cls.settings.kops_finalizer
        
        if finalizer != cls.settings.kops_finalizer:
            cls.settings.kops_finalizer = finalizer
        
        storage_prefix = storage_prefix if storage_prefix is not None \
            else cls.settings.kops_prefix
        persistent_key = persistent_key if persistent_key is not None \
            else cls.settings.kops_persistent_key
        
        if persistent_key != cls.settings.kops_persistent_key:
            cls.settings.kops_persistent_key = persistent_key
        
        error_delays = error_delays if error_delays is not None \
            else [10, 20, 30]
        kopf_name = kopf_name if kopf_name is not None \
            else cls.settings.kopf_name
        app_name = app_name if app_name is not None \
            else cls.settings.app_name
        
        _logger = _logger if _logger is not None \
            else logger
        
        if startup_functions is not None:
            for func in startup_functions:
                cls.add_function(func, EventType.startup)
        
        if shutdown_functions is not None:
            for func in shutdown_functions:
                cls.add_function(func, EventType.shutdown)

        
        # @kopf.on.startup()
        # async def configure(settings: kopf.OperatorSettings, logger: logging.Logger, **kwargs):
        
        @kopf.on.startup()
        async def configure(settings: kopf.OperatorSettings, **kwargs):
            if _settings is not None:
                settings = _settings
            if enable_event_logging is False:
                settings.posting.enabled = enable_event_logging
                _logger.info(f'Kopf Events Enabled: {enable_event_logging}')
            if event_logging_level is not None:
                # elif enable_event_logging is True:
                settings.posting.level = logging.getLevelName(event_logging_level.upper())
                _logger.info(f'Kopf Events Logging Level: {event_logging_level}')

            settings.persistence.finalizer = finalizer
            settings.persistence.progress_storage = kopf.SmartProgressStorage(prefix = storage_prefix)
            settings.persistence.diffbase_storage = kopf.AnnotationsDiffBaseStorage(
                prefix = storage_prefix,
                key = persistent_key,
            )
            settings.batching.error_delays = error_delays
            if multi_pods:
                settings.peering.priority = random.randint(0, 32767)
                settings.peering.stealth = True
                if kwargs.get('peering_name'):
                    settings.peering.name = kwargs.get('peering_name')

            # if cls.settings.kopf_enable_event_logging is False:
            #     settings.posting.enabled = cls.settings.kopf_enable_event_logging
            #     logger.info(f'Kopf Events Enabled: {cls.settings.kopf_enable_event_logging}')
            # elif cls.settings.kopf_enable_event_logging is True:
            #     settings.posting.level = logging.getLevelName(cls.settings.kopf_event_logging_level.upper())
            #     logger.info(f'Kopf Events Logging Level: {cls.settings.kopf_enable_event_logging}')

            # settings.persistence.finalizer = cls.settings.kops_finalizer
            # settings.persistence.progress_storage = kopf.SmartProgressStorage(prefix = cls.settings.kops_prefix)
            # settings.persistence.diffbase_storage = kopf.AnnotationsDiffBaseStorage(
            #     prefix = cls.settings.kops_prefix,
            #     key = cls.settings.kops_persistent_key,
            # )
            # settings.batching.error_delays = [10, 20, 30]

            _logger.info(f'Starting Kopf: {kopf_name} {app_name} @ {cls.settings.build_id}')
            await cls.aset_k8_config()
            if cls._startup_functions:
                _logger.info('Running Startup Functions')
                await cls.run_startup_functions()
            _logger.info('Completed Kopf Startup')

        @kopf.on.login()
        async def login_fn(**kwargs):
            return (
                kopf.login_with_service_account(**kwargs)
                or kopf.login_with_kubeconfig(**kwargs)
                if cls.settings.in_k8s
                else kopf.login_via_client(**kwargs)
            )
        

class BaseKOpsClient(metaclass = KOpsClientMeta):
    """
    Provides a base client for interacting with Kubernetes.
    """
    pass


"""
Cached Methods for Efficient Retrieval
"""

def get_namespaces(**kwargs) -> List[t.V1Namespace]:
    namespaces: t.V1NamespaceList = BaseKOpsClient.core_v1.list_namespace(**kwargs)
    return list(namespaces.items)

@cached(ttl = 5, serializer = DillSerializer)
async def aget_namespaces(**kwargs) -> List[at.V1Namespace]:
    namespaces: at.V1NamespaceList = await BaseKOpsClient.acore_v1.list_namespace(**kwargs)
    return list(namespaces.items)


"""
Nodes
"""

# @cached(ttl = 60, serializer = DillSerializer)
def get_node_name(name: Optional[str] = None, host_ip: Optional[str] = None, **kwargs) -> str:
    # not sure how this would work with diff clusters
    assert name or host_ip, 'Must provide either name or host_ip'
    if host_ip: name = 'ip-' + host_ip.replace('.', '-') + '.ec2.internal'
    return name

def get_nodes(**kwargs) -> List[t.V1Node]:
    nodes: t.V1NodeList = BaseKOpsClient.core_v1.list_node(**kwargs)
    return list(nodes.items)

@cached(ttl = 5, serializer = DillSerializer)
async def aget_nodes(**kwargs) -> List[at.V1Node]:
    nodes: at.V1NodeList = await BaseKOpsClient.acore_v1.list_node(**kwargs)
    return list(nodes.items)


def get_node(name: Optional[str] = None, host_ip: Optional[str] = None, **kwargs) -> t.V1Node:
    name = get_node_name(name = name, host_ip = host_ip)
    return BaseKOpsClient.core_v1.read_node(name = name, **kwargs)

@cached(ttl = 5, serializer = DillSerializer)
async def aget_node(name: Optional[str] = None, host_ip: Optional[str] = None, **kwargs) -> at.V1Node:
    name = get_node_name(name = name, host_ip = host_ip)
    return await BaseKOpsClient.acore_v1.read_node(name = name, **kwargs)


"""
Pods
"""
def get_pods(namespace: Optional[str] = None, **kwargs) -> List[t.V1Pod]:
    pods: t.V1PodList = BaseKOpsClient.core_v1.list_pod_for_all_namespaces(**kwargs)
    pods = list(pods.items)
    if namespace: pods = [pod for pod in pods if pod.metadata.namespace == namespace]
    return pods

@cached(ttl = 5, serializer = DillSerializer)
async def aget_pods(namespace: Optional[str] = None, **kwargs) -> List[at.V1Pod]:
    pods: at.V1PodList = await BaseKOpsClient.acore_v1.list_pod_for_all_namespaces(**kwargs)
    pods = list(pods.items)
    if namespace: pods = [pod for pod in pods if pod.metadata.namespace == namespace]
    return pods


def get_pod(name: str, namespace: Optional[str] = None, **kwargs) -> t.V1Pod:
    if name and namespace:
        return BaseKOpsClient.core_v1.read_namespaced_pod(name = name, namespace = namespace, **kwargs)
    pods = get_pods(namespace = namespace, **kwargs)
    for pod in pods:
        if pod.metadata.name == name:
            return pod
    raise ValueError(f'Pod {name} not found')

@cached(ttl = 5, serializer = DillSerializer)
async def aget_pod(name: str, namespace: Optional[str] = None, **kwargs) -> at.V1Pod:
    if name and namespace:
        return await BaseKOpsClient.acore_v1.read_namespaced_pod(name = name, namespace = namespace, **kwargs)
    pods = await aget_pods(namespace = namespace, **kwargs)
    for pod in pods:
        if pod.metadata.name == name:
            return pod
    raise ValueError(f'Pod {name} not found')


@cached(ttl = 5, serializer = DillSerializer)
async def aget_pods_on_node(name: str, host_ip: Optional[str] = None, namespace: Optional[str] = None, **kwargs) -> List[at.V1Pod]:
    pods = await aget_pods(namespace = namespace, **kwargs)
    host_ip = get_node_name(name = name, host_ip = host_ip)
    return [pod for pod in pods if pod.status.host_ip and pod.status.host_ip == host_ip]

@cached(ttl = 5, serializer = DillSerializer)
async def aget_claimnames_from_pods_on_node(name: str, host_ip: Optional[str] = None, namespace: Optional[str] = None, **kwargs) -> List[str]:
    """
    Returns a list of PVC names from pods on a node
    """
    pods = await aget_pods_on_node(name = name, host_ip = host_ip, namespace = namespace, **kwargs)
    claim_names = []
    for pod in pods:
        claim_names.extend(
            volume.persistent_volume_claim.claim_name
            for volume in pod.spec.volumes
            if volume.persistent_volume_claim
        )
    return claim_names


def run_pod_command(
    name: str, 
    namespace: str,
    command: Union[str, List[str]],
    container: Optional[str] = None,
    stderr: Optional[bool] = True,
    stdin: Optional[bool] = True,
    stdout: Optional[bool] = True,
    tty: Optional[bool] = False,
    ignore_error: bool = True, 
    **kwargs
) -> str:
    """
    Runs a command in the pod and returns the command
    name = 'name_example' # str | name of the PodExecOptions
    namespace = 'namespace_example' # str | object name and auth scope, such as for teams and projects
    command = 'command_example' # str | Command is the remote command to execute. argv array. Not executed within a shell. (optional)
    container = 'container_example' # str | Container in which to execute the command. Defaults to only container if there is only one container in the pod. (optional)
    stderr = True # bool | Redirect the standard error stream of the pod for this call. (optional)
    stdin = True # bool | Redirect the standard input stream of the pod for this call. Defaults to false. (optional)
    stdout = True # bool | Redirect the standard output stream of the pod for this call. (optional)
    tty = True # bool | TTY if true indicates that a tty will be allocated for the exec call. Defaults to false. (optional)

    """
    if isinstance(command, str):
        command = command.split(' ')
    try:
        return BaseKOpsClient.core_v1_ws.connect_post_namespaced_pod_exec(
            name = name, 
            namespace = namespace, 
            command = command,
            container = container,
            stderr = stderr,
            stdin = stdin,
            stdout = stdout,
            tty = tty
        )
    except Exception as e:
        if not ignore_error:
            logger.warning(f'[{namespace}/{name}] Exec Status Failed: {e}')
        return None

async def arun_pod_command(
    name: str, 
    namespace: str,
    command: Union[str, List[str]],
    container: Optional[str] = None,
    stderr: Optional[bool] = True,
    stdin: Optional[bool] = True,
    stdout: Optional[bool] = True,
    tty: Optional[bool] = False,
    ignore_error: bool = True, 
    **kwargs
) -> str:
    """
    Runs a command in the pod and returns the command
    name = 'name_example' # str | name of the PodExecOptions
    namespace = 'namespace_example' # str | object name and auth scope, such as for teams and projects
    command = 'command_example' # str | Command is the remote command to execute. argv array. Not executed within a shell. (optional)
    container = 'container_example' # str | Container in which to execute the command. Defaults to only container if there is only one container in the pod. (optional)
    stderr = True # bool | Redirect the standard error stream of the pod for this call. (optional)
    stdin = True # bool | Redirect the standard input stream of the pod for this call. Defaults to false. (optional)
    stdout = True # bool | Redirect the standard output stream of the pod for this call. (optional)
    tty = True # bool | TTY if true indicates that a tty will be allocated for the exec call. Defaults to false. (optional)

    """
    if isinstance(command, str):
        command = command.split(' ')
    
    try:
        return await BaseKOpsClient.acore_v1_ws.connect_post_namespaced_pod_exec(
            name = name, 
            namespace = namespace, 
            command = command,
            container = container,
            stderr = stderr,
            stdin = stdin,
            stdout = stdout,
            tty = tty
        )
    except Exception as e:
        if not ignore_error:
            logger.warning(f'[{namespace}/{name}] Exec Status Failed: {e}')
        return None

"""
PVCs
"""

@cached(ttl = 5, serializer = DillSerializer)
async def aget_pvcs(namespace: Optional[str] = None, **kwargs) -> List[at.V1PersistentVolumeClaim]:
    pvcs: at.V1PersistentVolumeClaimList = await BaseKOpsClient.acore_v1.list_persistent_volume_claim_for_all_namespaces(**kwargs)
    pvcs = list(pvcs.items)
    if namespace: pvcs = [pvc for pvc in pvcs if pvc.metadata.namespace == namespace]
    return pvcs


@cached(ttl = 5, serializer = DillSerializer)
async def aget_pvcs_from_claim_names(claim_names: List[str], namespace: Optional[str] = None, **kwargs) -> List[at.V1PersistentVolumeClaim]:
    """
    Used to filter down PVCs that are attached to a list of pods
    """
    pvcs = await aget_pvcs(namespace = namespace, **kwargs)
    return [pvc for pvc in pvcs if pvc.metadata.name in claim_names]


@cached(ttl = 5, serializer = DillSerializer)
async def aget_pvcs_with_storageclass(storageclass: str, namespace: Optional[str] = None, name: Optional[str] = None,  **kwargs) -> Union[List[at.V1PersistentVolumeClaim], at.V1PersistentVolumeClaim]:
    pvcs = await aget_pvcs(namespace = namespace, **kwargs)
    pvcs = [pvc for pvc in pvcs if pvc.spec.storage_class_name == storageclass]
    if name:
        for pvc in pvcs:
            if pvc.metadata.name == name:
                return pvc
        raise ValueError(f'PVC {name} not found')
    return pvcs

@cached(ttl = 5, serializer = DillSerializer)
async def aget_pvcs_with_storageclasses(storageclasses: List[str], namespace: Optional[str] = None, **kwargs) -> List[at.V1PersistentVolumeClaim]:
    pvcs = await aget_pvcs(namespace = namespace, **kwargs)
    return [pvc for pvc in pvcs if pvc.spec.storage_class_name in storageclasses]

@cached(ttl = 5, serializer = DillSerializer)
async def aget_pvc(name: str, namespace: Optional[str] = None, storageclass: Optional[str] = None, **kwargs) -> at.V1PersistentVolumeClaim:
    if storageclass:
        return await aget_pvcs_with_storageclass(storageclass = storageclass, name = name, namespace = namespace, **kwargs)
    if name and namespace:
        return await BaseKOpsClient.acore_v1.read_namespaced_persistent_volume_claim(name = name, namespace = namespace, **kwargs)
    pvcs = await aget_pvcs(namespace = namespace, **kwargs)
    for pvc in pvcs:
        if pvc.metadata.name == name:
            return pvc
    raise ValueError(f'PVC {name} not found')


"""
Combination Methods
"""

@cached(ttl = 5, serializer = DillSerializer)
async def aget_pods_with_pvc(namespace: Optional[str] = None, **kwargs) -> List[at.V1Pod]:
    """
    Returns all pods that have a PVC
    """
    pods = await aget_pods(namespace = namespace, **kwargs)
    return [
        pod
        for pod in pods
        if pod.spec.volumes
        and any(v.persistent_volume_claim for v in pod.spec.volumes)
    ]

@cached(ttl = 5, serializer = DillSerializer)
async def aget_pvcs_on_node(name: Optional[str] = None, host_ip: Optional[str] = None, namespace: Optional[str] = None, **kwargs) -> List[at.V1PersistentVolumeClaim]:
    """
    Returns all PVCs that are on a given node
    """
    claim_names = await aget_claimnames_from_pods_on_node(name = name, host_ip = host_ip, namespace = namespace, **kwargs)
    return await aget_pvcs_from_claim_names(claim_names = claim_names, namespace = namespace, **kwargs)

@cached(ttl = 5, serializer = DillSerializer)
async def get_pvcs_on_node_with_storageclasses(storageclasses: List[str], name: Optional[str] = None, host_ip: Optional[str] = None, namespace: Optional[str] = None, **kwargs) -> List[at.V1PersistentVolumeClaim]:
    """
    Returns all PVCs that are on a given node that have a given storageclass
    """
    pvcs = await aget_pvcs_on_node(name = name, host_ip = host_ip, namespace = namespace, **kwargs)
    return [pvc for pvc in pvcs if pvc.spec.storage_class_name in storageclasses]


class KOpsClient(BaseKOpsClient):

    @classmethod
    def get_namespaces(cls, **kwargs) -> List[str]:
        return get_namespaces(**kwargs)
    
    @classmethod
    async def aget_namespaces(cls, **kwargs) -> List[str]:
        return await aget_namespaces(**kwargs)
    
    @classmethod
    def get_nodes(cls, **kwargs) -> List[t.V1Node]:
        return get_nodes(**kwargs)

    @classmethod
    async def aget_nodes(cls, **kwargs) -> List[at.V1Node]:
        return await aget_nodes(**kwargs)
    
    @classmethod
    def get_pods(cls, namespace: Optional[str] = None, **kwargs) -> List[t.V1Pod]:
        return get_pods(namespace = namespace, **kwargs)
    
    @classmethod
    async def aget_pods(cls, namespace: Optional[str] = None, **kwargs) -> List[at.V1Pod]:
        return await aget_pods(namespace = namespace, **kwargs)
    
    @classmethod
    def run_command(
        cls,
        name: str, 
        namespace: str,
        command: Union[str, List[str]],
        container: Optional[str] = None,
        stderr: Optional[bool] = True,
        stdin: Optional[bool] = True,
        stdout: Optional[bool] = True,
        tty: Optional[bool] = False,
        ignore_error: bool = True, 
        **kwargs
    ) -> str:
        return run_pod_command(
            name = name, 
            namespace = namespace,
            command = command,
            container = container,
            stderr = stderr,
            stdin = stdin,
            stdout = stdout,
            tty = tty,
            ignore_error = ignore_error, 
            **kwargs
        )
    
    @classmethod
    async def arun_command(
        cls,
        name: str, 
        namespace: str,
        command: Union[str, List[str]],
        container: Optional[str] = None,
        stderr: Optional[bool] = True,
        stdin: Optional[bool] = True,
        stdout: Optional[bool] = True,
        tty: Optional[bool] = False,
        ignore_error: bool = True, 
        **kwargs
    ) -> str:
        return await arun_pod_command(
            name = name, 
            namespace = namespace,
            command = command,
            container = container,
            stderr = stderr,
            stdin = stdin,
            stdout = stdout,
            tty = tty,
            ignore_error = ignore_error, 
            **kwargs
        )



class SqlTimeout(Exception):
    "Database timeout expired."


class UnknownFileWarning(UserWarning):
    "Warning used by Store.check for unknown files."


class EmptyDirWarning(UserWarning):
    "Warning used by Store.check for empty directories."



"""
Compression Utilities
"""

from typing import Union
from .base import BaseCompress
from .gz import GzipCompress
from .lz import Lz4Compress
from .zb import ZLibCompress
from .ztd import ZStdCompress

CompressionT = Union[BaseCompress, GzipCompress]

class Compress:

    gzip = GzipCompress
    zstd = ZStdCompress
    zlib = ZLibCompress
    lz4 = Lz4Compress
    



"""
Import Handler for sqlalchemy
"""

from lazyops.utils.imports import resolve_missing, require_missing_wrapper

try:
    import sqlalchemy
    _sqlalchemy_available = True
except ImportError:
    sqlalchemy = object
    _sqlalchemy_available = False

try:
    import asyncpg
    _asyncpg_available = True
except ImportError:
    asyncpg = object
    _asyncpg_available = False

try:
    import psycopg2
    _psycopg2_available = True
except ImportError:
    psycopg2 = object
    _psycopg2_available = False

try:
    import sqlalchemy_json
    _sqlalchemy_json_available = True
except ImportError:
    sqlalchemy_json = object
    _sqlalchemy_json_available = False

def resolve_sqlalchemy(
    required: bool = False,
):
    """
    Ensures that `sqlalchemy` is available
    """
    global sqlalchemy, _sqlalchemy_available
    if not _sqlalchemy_available:
        resolve_missing('sqlalchemy', required = required)
        import sqlalchemy
        _sqlalchemy_available = True
        globals()['sqlalchemy'] = sqlalchemy

def resolve_asyncpg(
    required: bool = False,
):
    """
    Ensures that `asyncpg` is available
    """
    global asyncpg, _asyncpg_available
    if not _asyncpg_available:
        resolve_missing('asyncpg', required = required)
        import asyncpg
        _asyncpg_available = True
        globals()['asyncpg'] = asyncpg

def resolve_psycopg2(
    required: bool = False,
):
    """
    Ensures that `psycopg2` is available
    """
    global psycopg2, _psycopg2_available
    if not _psycopg2_available:
        resolve_missing('psycopg2', 'psycopg2-binary', required = required)
        import psycopg2
        _psycopg2_available = True
        globals()['psycopg2'] = psycopg2

def resolve_sqlalchemy_json(
    required: bool = False,
):
    """
    Ensures that `sqlalchemy_json` is available
    """
    global sqlalchemy_json, _sqlalchemy_json_available
    if not _sqlalchemy_json_available:
        resolve_missing('sqlalchemy_json', required = required)
        import sqlalchemy_json
        _sqlalchemy_json_available = True
        globals()['sqlalchemy_json'] = sqlalchemy_json

def require_sqlalchemy(
    required: bool = False,
):
    """
    Wrapper for `resolve_sqlalchemy` that can be used as a decorator
    """
    def decorator(func):
        return require_missing_wrapper(
            resolver = resolve_sqlalchemy, 
            func = func, 
            required = required
        )
    return decorator

def resolve_sql(
    required: bool = True,
    require_sqlalchemy: bool = True,
    require_psycopg2: bool = False,
    require_asyncpg: bool = False,
    require_sqlalchemy_json: bool = True,
):
    """
    Ensures that `sqlalchemy` is available
    """
    if require_sqlalchemy:
        resolve_sqlalchemy(required = required)
    if require_psycopg2:
        resolve_psycopg2(required = required)
    if require_asyncpg:
        resolve_asyncpg(required = required)
    if require_sqlalchemy_json:
        resolve_sqlalchemy_json(required = required)

def require_sql(
    required: bool = True,
    require_sqlalchemy: bool = True,
    require_psycopg2: bool = False,
    require_asyncpg: bool = False,
    require_sqlalchemy_json: bool = True,
):
    """
    Wrapper for `resolve_sqlalchemy` that can be used as a decorator
    """
    def decorator(func):
        return require_missing_wrapper(
            resolver = resolve_sql, 
            func = func, 
            required = required,
            require_sqlalchemy = require_sqlalchemy,
            require_psycopg2 = require_psycopg2,
            require_asyncpg = require_asyncpg,
            require_sqlalchemy_json = require_sqlalchemy_json,
        )
    return decorator


import pickle
from typing import Any, Dict, Optional, Union, Type
from lazyops.utils.lazy import lazy_import
from .base import BinaryBaseSerializer, ObjectValue, SchemaType, BaseModel, logger

try:
    import cloudpickle
    _cloudpicke_available = True
except ImportError:
    _cloudpicke_available = False

try:
    import dill
    _dill_available = True
except ImportError:
    _dill_available = False

if _cloudpicke_available:
    default_pickle = cloudpickle
elif _dill_available:
    default_pickle = dill
else:
    default_pickle = pickle

class PickleSerializer(BinaryBaseSerializer):
    name: Optional[str] = "pickle"
    encoding: Optional[str] = "utf-8"
    picklelib: Any = default_pickle

    def __init__(
        self, 
        picklelib: Optional[Union[str, Any]] = None,
        compression: Optional[str] = None,
        compression_level: int | None = None, 
        encoding: str | None = None, 
        **kwargs
    ):
        super().__init__(compression, compression_level, encoding, **kwargs)
        if picklelib is not None:
            if isinstance(picklelib, str):
                picklelib = lazy_import(picklelib)
            assert hasattr(picklelib, "dumps") and hasattr(picklelib, "loads"), f"Invalid Pickle Library: {picklelib}"
            self.picklelib = picklelib
        self.picklelib_name = self.picklelib.__name__
    
    def encode_value(self, value: Union[Any, SchemaType], **kwargs) -> bytes:
        """
        Encode the value with the Pickle Library
        """
        try:
            return self.picklelib.dumps(value, **kwargs)
        except Exception as e:
            logger.info(f'Error Encoding Value: |r|({type(value)}) {e}|e| {value}', colored = True, prefix = self.picklelib_name)
            if self.raise_errors: raise e
        return None

    def decode_value(self, value: bytes, **kwargs) -> Union[SchemaType, Dict, Any]:
        """
        Decode the value with the Pickle Library
        """
        try:
            if self.picklelib_name == 'cloudpickle':
                if 'encoding' not in kwargs:
                    kwargs['encoding'] = self.encoding
                if 'fix_imports' not in kwargs:
                    kwargs['fix_imports'] = False
            return self.picklelib.loads(value, **kwargs)
        except Exception as e:
            logger.info(f'Error Decoding Value: |r|({type(value)}) {e}|e| {value}', colored = True, prefix = self.picklelib_name)
            if self.raise_errors: raise e
        return None


from .base import BaseStatefulBackend
from .local import LocalStatefulBackend
from .redis import RedisStatefulBackend

from typing import Union

StatefulBackendT = Union[LocalStatefulBackend, RedisStatefulBackend, BaseStatefulBackend]


"""
Preconfigured Worker Configurations
"""

import contextlib
from lazyops.types.models import BaseSettings
from lazyops.utils.helpers import import_function
from typing import List, Optional, Dict, Any, Callable, Union, TYPE_CHECKING


if TYPE_CHECKING:
    with contextlib.suppress(ImportError):
        from aiokeydb import Worker, TaskQueue



class WorkerSettings(BaseSettings):
    """
    Worker Queue Settings
    """
    enabled: bool = True

    queue_enabled: bool = True # if True, enables the queue
    queue_name: Optional[str] = None # the queue for the local worker. This will be set on init from main settings
    
    queue_shared_prefix: Optional[str] = "cluster"
    queue_db_id: Optional[int] = 3

    worker_cls: Optional[str] = None
    worker_serializer: Optional[str] = None

    socket_timeout: Optional[float] = 120.0
    socket_connect_timeout: Optional[float] = 120.0
    connection_timeout: Optional[int] = 600
    socket_keepalive: Optional[bool] = True
    retry_on_timeout: Optional[bool] = True
    health_check_interval: Optional[int] = 15
    is_leader_process: Optional[bool] = None


    class Config:
        case_sensitive = False
        env_prefix = "WORKER_"

    
    def get_queue(
        self, 
        name: str, 
        db_id: Optional[int] = None, 
        shared: Optional[bool] = False,
        serializer: Optional[str] = None,
        is_leader_process: Optional[bool] = None,
        **kwargs
    ) -> 'TaskQueue':
        """
        Returns a Task Queue

        Args:
            name (str): The name of the queue to use
            db_id (Optional[int]): The db id to use. If None, uses the default
            shared (Optional[bool]): If True, uses the shared prefix. If False, does not add a prefix
            kwargs: Any additional kwargs to pass to the TaskQueue

        """
        from aiokeydb import TaskQueue
        queue_name = f'{self.queue_shared_prefix}.{name}' if shared else name
        if serializer is not None: 
            from aiokeydb.serializers import SerializerType
            kwargs['serializer'] = SerializerType(serializer).get_serializer()
        
        is_leader_process = is_leader_process if is_leader_process is not None else self.is_leader_process
        return TaskQueue(
            queue_name = queue_name,
            db = db_id or self.queue_db_id,
            is_leader_process = is_leader_process,
            **kwargs
        )
    
    def get_worker(
        self, 
        worker_module: Union[str, Callable], 
        name: Optional[str] = None, 
        worker_name: Optional[str] = None,
        queue: Optional['TaskQueue'] = None,
        queue_kwargs: Optional[Dict[str, Any]] = None,
        is_leader_process: Optional[bool] = None,
        **kwargs,
    ) -> 'Worker':
        """
        Returns a Worker from a Task Queue

        Args:
            worker_module (Union[str, Callable]): The worker class to use. Imports the class
            name (Optional[str]): The name of the queue to use. If None, uses the first part of the worker_cls name
            worker_name (Optional[str], optional): The name of the worker. Defaults to None.
            queue (Optional['TaskQueue'], optional): The queue to use. Defaults to None.
            queue_kwargs (Optional[Dict[str, Any]], optional): The kwargs to pass to the queue. Defaults to None.
            **kwargs: The kwargs to pass to the worker

        """
        worker_cls: 'Worker' = import_function(worker_module)
        if name is None: name = worker_module.split('.')[0]
        queue_kwargs = queue_kwargs or {}
        is_leader_process = is_leader_process if is_leader_process is not None else self.is_leader_process
        try:
            return worker_cls(queue = queue or self.get_queue(name, **queue_kwargs), name = worker_name or name, is_leader_process = is_leader_process, **kwargs)
        except Exception:
            return worker_cls(queue = queue or self.get_queue(name, **queue_kwargs), name = worker_name or name, **kwargs)



"""
Import Handler for `fileio`
"""

from lazyops.utils.imports import resolve_missing, require_missing_wrapper

try:
    from fileio import File, FileLike
    _fileio_available = True
except ImportError:
    from os import PathLike
    File = object
    FileLike = PathLike
    _fileio_available = False

def resolve_fileio(
    required: bool = False,
):
    """
    Ensures that `fileio` is available
    """
    global fileio, _fileio_available
    if not _fileio_available:
        resolve_missing('fileio', 'file-io', required = required)
        import fileio
        _fileio_available = True
        globals()['fileio'] = fileio


def require_fileio(
    required: bool = False,
):
    """
    Wrapper for `resolve_fileio` that can be used as a decorator
    """
    def decorator(func):
        return require_missing_wrapper(
            resolver = resolve_fileio, 
            func = func, 
            required = required
        )
    return decorator



"""
Import Handler for dateparser and pytz
"""


try:
    import dateparser
    import pytz
    _dateparser_available = True
except ImportError:
    dateparser = object
    pytz = object
    _dateparser_available = False

def resolve_dateparser(
    required: bool = False,
):
    """
    Ensures that `dateparser` is available
    """
    global dateparser, pytz, _dateparser_available
    if not _dateparser_available:
        from lazyops.utils.imports import resolve_missing
        resolve_missing('dateparser', required = required)
        import dateparser
        import pytz
        _dateparser_available = True
        globals()['dateparser'] = dateparser
        globals()['pytz'] = pytz


def require_dateparser(
    required: bool = False,
):
    """
    Wrapper for `resolve_dateparser` that can be used as a decorator
    """
    def decorator(func):
        from lazyops.utils.imports import require_missing_wrapper
        return require_missing_wrapper(
            resolver = resolve_dateparser, 
            func = func, 
            required = required
        )
    return decorator


from __future__ import annotations

"""
A File Conversion based on PDFtoText

- Requires pdftotext to be installed
"""
import os
import asyncio
import subprocess
from pathlib import Path
from ..base import (
    BaseConverter, 
    OutputType, 
    InputContentType, 
    InputPathType, 
    InputSourceType, 
    InvalidSourceError, 
    InvalidTargetError
)

from typing import Any, Union, Optional, Type, Iterable, Callable, Dict, List, Tuple, TypeVar


class PDFtoTextConverter(BaseConverter):

    name: str = 'pdftotext'
    source: str = 'pdf'
    targets: List[str] = ['txt', 'text']
    async_enabled: bool = True


    def validate_enabled(self, **kwargs) -> bool:
        """
        Validate whether the converter is enabled
        """
        from shutil import which

        return which('pdftotext') is not None

    
    def _convert_source_to_target(
        self,
        source: InputSourceType,
        target: Optional[str] = '.txt', 

        source_filename: Optional[str] = None,
        target_output: Optional[Any] = None,

        raise_errors: Optional[bool] = False,
        **kwargs,
    ) -> OutputType:
        """
        Convert the source to the target

        source: /path/to/file.pdf
        target: '.txt'
        """
        if not self.is_valid_target_content_type(target):
            raise InvalidTargetError(self, target)

        target_output_path = Path(target_output) if target_output else None
        if target_output_path and target_output_path.is_dir():
            if not source_filename and isinstance(source, str):
                target_filename = Path(source).with_suffix(target).name
            else:
                target_filename = f'output{target}'
            target_output_path = target_output_path.joinpath(target_filename)
        
        source_file = self.convert_file_input_to_file(source, source_filename, make_temp = True)
        cmd = f'cat "{source_file.as_posix()}" | pdftotext -layout -nopgbrk -eol unix -colspacing 0.7 -y 58 -x 0 -H 741 -W 596 - -'
        process = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        try:
            stdout, stderr = process.communicate()
            stdout = stdout.decode('utf-8')
            os.unlink(source_file.as_posix())
            if target_output_path:
                target_output_path.parent.mkdir(parents=True, exist_ok=True)
                target_output_path.write_text(stdout)
                return target_output_path
            return stdout
        except Exception as e:
            stderr = stderr.decode('utf-8')
            self.logger.error(f'Error in pdftotext: {stderr}: {e}')
            os.unlink(source_file.as_posix())
            if raise_errors: raise e
            return None


    async def _aconvert_source_to_target(
        self,
        source: InputSourceType,
        target: Optional[str] = '.txt', 

        source_filename: Optional[str] = None,
        target_output: Optional[Any] = None,

        raise_errors: Optional[bool] = False,
        **kwargs,
    ) -> OutputType:
        """
        Convert the source to the target

        source: /path/to/file.pdf
        target: '.txt'
        """
        if not self.is_valid_target_content_type(target):
            raise InvalidTargetError(self, target)

        target_output_path = Path(target_output) if target_output else None
        if target_output_path and target_output_path.is_dir():
            if not source_filename and isinstance(source, str):
                target_filename = Path(source).with_suffix(target).name
            else:
                target_filename = f'output{target}'
            target_output_path = target_output_path.joinpath(target_filename)
        
        source_file = self.convert_file_input_to_file(source, source_filename, make_temp = True)
        cmd = f'cat "{source_file.as_posix()}" | pdftotext -layout -nopgbrk -eol unix -colspacing 0.7 -y 58 -x 0 -H 741 -W 596 - -'
        process = await asyncio.subprocess.create_subprocess_shell(cmd, shell=True, stdout=asyncio.subprocess.PIPE, stderr=asyncio.subprocess.PIPE)
        try:
            stdout, stderr = await process.communicate()
            stdout = stdout.decode('utf-8')
            os.unlink(source_file.as_posix())
            if target_output_path:
                target_output_path.parent.mkdir(parents=True, exist_ok=True)
                target_output_path.write_text(stdout)
                return target_output_path
            return stdout
        except Exception as e:
            stderr = stderr.decode('utf-8')
            self.logger.error(f'Error in pdftotext: {stderr}: {e}')
            os.unlink(source_file.as_posix())
            if raise_errors: raise e
            return None
            



# Asyncz - Forked from [FastAsync](https://github.com/thebowenfeng/FastAsync/)


from .errors import AsyncTaskException
from .types import (
    WaitType,
    ErrorHandling,
    TaskStatus,
    AsyncTask, 
    global_task_manager
)
from .utils import (
    await_all,
    await_first,
    await_until
)

from typing import Callable

def asyncify(func: Callable):
    """
    Turn a function into an async task using asyncz task manager
    """
    def wrapper(*args, **kwargs):
        task = AsyncTask(func, *args, **kwargs)
        task.run()
        return task
    return wrapper


def set_max_threads(num : int):
    global_task_manager.max_threads(num)




from __future__ import annotations


import datetime
from lazyops.imports._dateparser import (
    dateparser, pytz, resolve_dateparser
)
from typing import Optional, List, Union

"""
General Datetime Utils
"""

tz_map = {
    'PST': 'US/Pacific', 
    'EST': 'US/Eastern',
    'EDT': 'US/Eastern',
    'CST': 'US/Central',
    'CDT': 'US/Central',
    'MST': 'US/Mountain',
    'MDT': 'US/Mountain',
    'PDT': 'US/Pacific',
    'AKST': 'US/Alaska',
    'AKDT': 'US/Alaska',
    'HST': 'US/Hawaii',
    'HAST': 'US/Hawaii',
    'HADT': 'US/Hawaii',
    'SST': 'US/Samoa',
    'SDT': 'US/Samoa',
    'CHST': 'Pacific/Guam',
    'CHDT': 'Pacific/Guam',
    
}

def convert_possible_date(
    dt: Optional[str] = None
) -> Optional[datetime.datetime]:
    """
    Converts a possible date string into a datetime
    """
    # Handles various date formats

    if not dt: return None
    dt = dt.strip()
    if not dt: return None
    
    resolve_dateparser(True)
    # print(f'WARNING: Could not parse date: {dt}')
    if dt[:2].isalpha():
        # Remove the timezone
        _tz = None
        if dt[-2:] not in {'AM', 'PM'}:
            dt, _tz = dt.rsplit(' ', 1)
            _tz = tz_map.get(_tz)

            if _tz: _tz = pytz.timezone(_tz)
        dt_val = datetime.datetime.strptime(dt, "%b %d, %Y %H:%M:%S %p")
        if _tz: dt_val = _tz.localize(dt_val).astimezone(pytz.utc)
        return dt_val

    if 'T' in dt:
        if '+' in dt:
            return datetime.datetime.strptime(dt, "%Y-%m-%dT%H:%M:%S%z")
        if '-' in dt[-7:]:
            # 2019-01-14T15:00:00-05:00 -> 2019-01-14T15:00:00
            dt = dt.rsplit('-', 1)[0]
            
        return datetime.datetime.strptime(dt, "%Y-%m-%dT%H:%M:%S")
    if ' ' in dt and '.' in dt:
        # "posted_date": "2023-04-17 21:57:14.756-04" -> "2023-04-17 21:57:14"
        dt = dt.split('.')[0]
        return datetime.datetime.strptime(dt, "%Y-%m-%d %H:%M:%S")
    
    if '-' in dt:
        return datetime.datetime.strptime(dt, "%Y-%m-%d")
    
    if '/' in dt:
        return datetime.datetime.strptime(dt, "%m/%d/%Y")
    # 20230418
    return datetime.datetime.strptime(dt, '%m%d%Y')

def convert_date(
    dt: Optional[str] = None
) -> Optional[datetime.datetime]:
    """
    Wrapper around dateparser.parse to handle some edge cases
    """
    if isinstance(dt, datetime.datetime): return dt
    
    try:
        return convert_possible_date(dt)
    except Exception as e:
        return dateparser.parse(dt)


def parse_datetime(
    dt: Optional[Union[str, List[str]]] = None, 
    timeaware: bool = True,
    tz: Optional[str] = None,
) -> Optional[Union[datetime.datetime, List[datetime.datetime]]]:
    """
    Parses the datetime string into a datetime
    """
    if not dt: return None
    resolve_dateparser(True)
    tz_info = None
    if tz:  tz_info = pytz.timezone(tz_map.get(tz.upper(), tz))
    elif timeaware: tz_info = datetime.timezone.utc

    if isinstance(dt, datetime.datetime):
        return dt.replace(tzinfo=tz_info) if tz_info \
            else dt

    if isinstance(dt, list): 
        dts = [convert_date(d) for d in dt]
        return [d.replace(tzinfo=tz_info) for d in dts] if tz_info else dts
    dt = convert_date(dt)
    if not dt: return None
    return dt.replace(tzinfo=tz_info) if tz_info \
        else dt
    

def parse_datetime_from_timestamp(
    timestamp: Optional[int],
) -> Optional[datetime.datetime]:
    """
    Parses the timestamp into a datetime

    Format: 1666699200000 -> 2022-12-24T00:00:00.000Z
    """
    if timestamp is None: return None
    return datetime.datetime.fromtimestamp(timestamp / 1000, tz = datetime.timezone.utc)


def get_current_datetime(
    tz: Optional[str] = None,
) -> datetime.datetime:
    """
    Gets the current datetime in the specified timezone
    """
    resolve_dateparser(True)
    if tz: tz = pytz.timezone(tz_map.get(tz.upper(), tz))
    return datetime.datetime.now(tz)


def is_expired_datetime(
    dt: datetime.datetime, 
    delta_days: Optional[int] = None, 
    now: Optional[datetime.datetime] = None,
    tz: Optional[str] = None,
) -> bool:
    """
    Checks if the datetime is expired
    """
    if not now: now = get_current_datetime(tz = tz)
    if not delta_days: return now > dt
    dt = dt + datetime.timedelta(days = delta_days)
    return now > dt